# -*- coding: utf-8 -*-

import time

from simmate.toolkit import Composition

from simmate.database import connect
from simmate.database.workflow_results import (
    EvolutionarySearch as SearchDatatable,
    StructureSource as SourceDatatable,
)

import simmate.toolkit.creators as creation_module
import simmate.toolkit.transformations.from_ase as transform_module
from simmate.toolkit.validators.fingerprint.pcrystalnn import (
    PartialCrystalNNFingerprint,
)

from typing import List, Union, Tuple
from simmate.workflow_engine import Workflow


class SearchEngine:
    """
    **WARNING** This search engine assumes you have properly configured
    Prefect Cloud and a cloud database backend (e.g. Postgres). In the future,
    we will accommodate local runs and other backends.

    This class is the entry point for predicting crystal structures with an
    evolutionary search algorithm.

    Once you initialize a SearchEngine with a composition (and any other
    optional parameters), you simply need to call the run() method to start
    the search:

    ``` python

        from simmate.toolkit.structure_prediction import SearchEngine

        search_engine = SearchEngine(
            composition="C4",
            labels=["WarWulf"],  # optional
            workflow_command="mpirun -n 8 vasp_std > vasp.out",  # optional
        )

        search.run()
    ```

    This class is only for creating (or continuing) searches. If you only want
    to view results, you should instead look directly at the database table for
    previously ran searches:

    ``` python

       from simmate.shortcuts import SearchResults

       search_results = SearchResults.objects.get(composition="Sr4 Si4 N8")

    ```

    If you start a separate terminal from where a search is running, you can
    actually use this table to view results WHILE the search is still running!

    You can also access the search results from your search_engine object. So
    once your search is finished or stopped, you can do:

    ``` python
       search_results = search_engine.search_datatable
    ```

    To better understand how to view/analyze results, please read the documentation
    for the SearchResults class.

    #### Alternative Codes & Softwares

    [USPEX](https://uspex-team.org/en),
    [XtalOpt](https://github.com/xtalopt/XtalOpt),
    [CALYPSO](http://www.calypso.cn/),
    [AIRSS](https://www.mtg.msm.cam.ac.uk/Codes/AIRSS),
    [ASE-GA](https://wiki.fysik.dtu.dk/ase/ase/ga/ga.html#module-ase.ga),
    [GASP](https://github.com/henniggroup/GASP-python)
    """

    def __init__(
        self,
        composition: Union[str, Composition],
        workflow: Union[str, Workflow] = "relaxation.vasp.staged",
        workflow_command: str = None,
        max_structures: int = 3000,
        limit_best_survival: int = 250,
        singleshot_sources: List[str] = [],
        nfirst_generation: int = 20,
        nsteadystate: int = 40,
        steadystate_sources: List[Tuple[float, str]] = [
            (0.30, "RandomSymStructure"),
            (0.30, "from_ase.Heredity"),
            (0.10, "from_ase.SoftMutation"),
            (0.10, "from_ase.MirrorMutation"),
            (0.05, "from_ase.LatticeStrain"),
            (0.05, "from_ase.RotationalMutation"),
            (0.05, "from_ase.AtomicPermutation"),
            (0.05, "from_ase.CoordinatePerturbation"),
        ],
        selector: str = "TruncatedSelection",
        # TODO: maybe use **workflow_run_kwargs...?
    ):
        """
        Sets up the search engine and its settings.

        #### Parameters

        - `composition`:
            The composition to run the evolutionary search for. Note that the
            number of sites is fixed to what is set here. (Ca2N vs Ca4N2)

        - `workflow`:
            The workflow to run all individuals through. Note, the result_database
            of this workflow will be treated as the individuals in this search.
            The default is "relaxation.vasp.staged"

        - `workflow_command`:
            The command that will be passed to the workflow.run() method.

        - `max_structures`:
            The maximum number of individuals that will be calculated before
            stopping the search. The default is 3000.

        - `limit_best_survival`:
            The search is stopped when the best individual remains unbeaten for
            this number of new individuals. The default is 250.

        - `singleshot_sources`:
            TODO: This is not implemented yet

        - `nfirst_generation`:
            No mutations or "child" individuals will be carried out until this
            number of individuals have been calculated. The default is 20.

        - `nsteadystate`:
            The total number of individuals from steady-state sources that will
            be running/submitted at any given time. The default is 40.

        - `steadystate_sources`:
            A list of tuples where each tuple is (percent, source). The percent
            determines the number of steady stage calculations that will be
            running for this at any given time. For example, 0.25 means
            0.25*40=10 individuals will be running/submitted at all times. The
            source can be from either the toolkit.creator or toolkit.transformations
            modules. Don't change this default unless you know what you're doing!

        - `selector`:
            The defualt method to use for choosing the parent individual(s). The
            default is TruncatedSelection.
        """

        # make sure we were givent a Composition object, and if not, we have
        # a string that should be converted to one.
        if not isinstance(composition, Composition):
            composition = Composition(composition)

        self.composition = composition
        self.workflow_command = workflow_command

        # TODO: consider grabbing these from the database so that we can update
        # them at any point.
        self.limit_best_survival = limit_best_survival
        self.max_structures = max_structures
        self.nsteadystate = nsteadystate
        self.nfirst_generation = nfirst_generation

        # Initialize the selector
        if selector == "TruncatedSelection":
            from simmate.toolkit.structure_prediction.evolution.selectors import (
                TruncatedSelection,
            )

            selector = TruncatedSelection()
        else:  # BUG
            raise Exception("We only support TruncatedSelection right now")
        self.selector = selector
        print(
            "The default parent selection for mutations will use "
            f"{self.selector.__class__.__name__}."
        )

        # Initialize the workflow if a string was given.
        # Otherwise we should already have a workflow class.
        if workflow == "relaxation.vasp.staged":
            from simmate.workflows.relaxation import Relaxation__Vasp__Staged

            self.workflow = Relaxation__Vasp__Staged

            # Point to the structure datatable that we'll pull from
            # BUG: For now, I assume its the results table of the workflow
            self.individuals_datatable = self.workflow.database_table._meta.get_field(
                "quality04relaxation"
            ).related_model  # gives quality04relaxation table
            self.calculation_datatable = self.workflow.database_table
            # BUG: these need to be merged and moved outside this if statement
        else:
            raise Exception(
                "Only `relaxation.vasp.staged` is supported in early testing"
            )
        print(
            f"All individuals will be evaulated through the "
            f"{self.workflow.name_full} workflow."
        )
        # BUG: I'll need to rewrite this in the future bc I don't really account
        # for other workflows yet. It would make sense that our workflow changes
        # as the search progresses (e.g. we incorporate DeePMD relaxation once ready)

        # Check if there is an existing search and grab it if so. Otherwise, add
        # the search entry to the DB.
        if SearchDatatable.objects.filter(composition=composition.formula).exists():
            self.search_datatable = SearchDatatable.objects.filter(
                composition=composition.formula,
            ).get()
            # TODO: update, add logs, compare... I need to decide what to do here.
            # also, run _check_stop_condition() to avoid unneccessary restart.
        else:
            self.search_datatable = SearchDatatable(
                composition=composition.formula,
                workflows=[
                    self.workflow.name_full
                ],  # as a list bc we can add new ones later
                individuals_datatable_str=self.individuals_datatable.__name__,
                max_structures=max_structures,
                limit_best_survival=limit_best_survival,
            )
            self.search_datatable.save()
        print(
            "To track the progress while this search runs, you can use the following"
            " in a separate python terminal:\n\n"
            "\tfrom simmate.shortcuts import SearchResults\n"
            f"\tsearch = SearchResults.objects.get(id={self.search_datatable.id})\n\n"
            f"So your search ID is {self.search_datatable.id}.\n"
            "View the documentation on the EvolutionarySearch datatable for more info."
        )

        # Grab the list of singleshot sources that have been ran before
        # and based off of that list, remove repeated sources
        past_singleshot_sources = (
            self.search_datatable.sources.filter(is_singleshot=True)
            .values_list("name", flat=True)
            .all()
        )
        singleshot_sources = [
            source
            for source in singleshot_sources
            if source not in past_singleshot_sources
        ]
        # BUG: what if I want to rerun any of these even though its been ran before?
        # An example would be rerunning substituitions when new structures are available

        # Initialize the single-shot sources
        self.singleshot_sources = []
        for source in singleshot_sources:

            # if we are given a string, we want initialize the class
            # otherwise it should be a class alreadys
            if type(source) == str:
                source = self._init_common_class(source)

            # and add it to our final list
            self.singleshot_sources.append(source)

        # Initialize the steady-state sources, which are given as a list of
        # (proportion, class/class_str, kwargs) for each. As we go through these,
        # we also collect the proportion list for them.
        self.steadystate_source_proportions = []
        self.steadystate_sources = []
        for proportion, source in steadystate_sources:

            # There are certain transformation sources that don't work for single-element
            # structures, so we check for this here and remove them.
            if len(composition.elements) == 1 and source in [
                "from_ase.AtomicPermutation"
            ]:
                print(
                    f"{source} is not possible with single-element structures."
                    " This is being removed from your steadystate_sources."
                )
                continue  # skips to next source

            # store proportion value
            self.steadystate_source_proportions.append(proportion)

            # if we are given a string, we want initialize the class
            # otherwise it should be a class already
            if type(source) == str:
                source = self._init_common_class(source)
            # and add it to our final list
            self.steadystate_sources.append(source)

        # Make sure the proportions sum to 1, otherwise scale them. We then convert
        # these to steady-state integers (and round to the nearest integer)
        sum_proportions = sum(self.steadystate_source_proportions)
        if sum_proportions != 1:
            print(
                "Warning: fractions for steady-state sources do not add to 1."
                "We have scaled all sources to equal one to fix this."
            )
            self.steadystate_source_proportions = [
                p / sum_proportions for p in self.steadystate_source_proportions
            ]
        # While these are percent values, we want specific counts. We convert to
        # those here.
        self.steadystate_source_counts = [
            int(p * nsteadystate) for p in self.steadystate_source_proportions
        ]

        # Throughout our search, we want to keep track of which workflows we've
        # submitted to prefect for each structure source as well as how long
        # we were holding a steady-state of submissions. We therefore keep
        # a log of Sources in our database -- where even if we've ran this search
        # before, we still want new entries for each.
        self.singleshot_sources_db = []
        for source in self.singleshot_sources:
            source_db = SourceDatatable(
                name=source.__class__.__name__,
                is_steadystate=False,
                is_singleshot=True,
                search=self.search_datatable,
            )
            source_db.save()
            self.singleshot_sources_db.append(source_db)
        self.steadystate_sources_db = []
        for source in self.steadystate_sources:
            source_db = SourceDatatable(
                name=source.__class__.__name__,
                is_steadystate=True,
                is_singleshot=False,
                search=self.search_datatable,
            )
            source_db.save()
            self.steadystate_sources_db.append(source_db)

        # Initialize the fingerprint database
        # For this we need to grab all previously calculated structures of this
        # compositon too pass in too.
        print("Generating fingerprints for past structures...")
        self.fingerprint_validator = PartialCrystalNNFingerprint(
            composition=composition,
            structure_pool=self.search_datatable.individuals,
        )
        # BUG: should we only do structures that were successfully calculated?
        # If not, there's a chance a structure fails because of something like a
        # crashed slurm job, but it's never submitted again...
        # OPTIMIZE: should we only do final structures? Or should I include input
        # structures and even all ionic steps as well...?
        print("Done.")

    def run(self, sleep_step=10):

        # See if the singleshot sources have been ran yet. For restarted calculations
        # this will likely not be needed (unless a new source was added)
        self._check_singleshot_sources()

        # this loop will go until I hit 'break' below
        while True:

            # TODO: maybe write summary files to csv...? This may be a mute
            # point because I expect we can follow along in the web UI in the future
            # To that end, I can add a "URL" property

            # Check the stop condition
            # If it is True, we can stop the calc.
            if self._check_stop_condition():
                break  # break out of the while loop
            # Otherwise, keep going!

            # update the fingerprint database with all of the completed structures
            self.fingerprint_validator.update_fingerprint_database()

            # Go through the running workflows and see if we need to submit
            # new ones to meet our steadystate target(s)
            self._check_steadystate_workflows()

            # TODO: Go through the triggered actions
            # self._check_triggered_actions()

            # To save our database load, sleep until we run checks again.
            # OPTIMIZE: ask Prefect if their is an equivalent to Dask's gather/wait
            # functions, so we know exactly when a workflow completes
            # https://docs.dask.org/en/stable/futures.html#waiting-on-futures
            print(f"Sleeping for {sleep_step} seconds before running checks again.")
            time.sleep(sleep_step)
        print("Stopping the search (remaining calcs will be left to finish).")

    def _check_stop_condition(self):

        # first see if we've hit our maximum limit for structures.
        # Note: because we are only looking at the results table, this is really
        # only counting the number of successfully calculated individuals.
        # Nothing is done to stop those that are still running or to count
        # structures that failed to be calculated
        # {f"{self.fitness_field}__isnull"=False} # when I allow other fitness fxns
        if self.search_datatable.individuals_completed.count() > self.max_structures:
            print(
                f"Maximum number of completed calculations hit (n={self.max_structures}."
            )
            return True
        # The 2nd stop condition is based on how long we've have the same
        # "best" individual. If the number of new individuals calculated (without
        # any becoming the new "best") is greater than limit_best_survival, then
        # we can stop the search.

        # grab the best individual for reference
        best = self.search_datatable.best_individual

        # We need this if-statement in case no structures have completed yet.
        if not best:
            return False

        # count the number of new individuals added AFTER the best one. If it is
        # more than limit_best_survival, we stop the search.
        num_new_structures_since_best = self.search_datatable.individuals.filter(
            # check energies to ensure we only count completed calculations
            energy_per_atom__gt=best.energy_per_atom,
            created_at__gte=best.created_at,
        ).count()
        if num_new_structures_since_best > self.limit_best_survival:
            print(
                f"Best individual has not changed after {self.limit_best_survival}"
                " new individuals added."
            )
            return True
        # If we reached this point, then we haven't hit a stop condition yet!
        return False

    def _check_singleshot_sources(self):
        print("Singleshot sources not implemented yet. Skipping this step.")

    def _check_steadystate_workflows(self):

        # we iterate through each steady-state source and check to see how many
        # jobs are still running for it. If it's less than the target steady-state,
        # then we need to submit more!
        for source, source_db, njobs_target in zip(
            self.steadystate_sources,
            self.steadystate_sources_db,
            self.steadystate_source_counts,
        ):
            # This loop says for the number of steady state runs we are short,
            # create that many new individuals! max(x,0) ensure we don't get a
            # negative value. A value of 0 means we are at steady-state and can
            # just skip this loop.
            for n in range(max(int(njobs_target - source_db.nprefect_flow_runs), 0)):

                # now we need to make a new individual and submit it!
                parent_ids, structure = self._make_new_structure(source)

                # sometimes we fail to make a structure with the source. In cases
                # like this, we warn the user, but just move on. This means
                # we will be short of our steady-state target. The warning for
                # this is done inside _make_new_structure
                if not structure:
                    break

                # submit the structure workflow
                # Note, we only pass the workflow_command if it's been supplied.
                extra_kwargs = (
                    {"command": self.workflow_command} if self.workflow_command else {}
                )
                flow_run_id = self.workflow.run_cloud(
                    structure=structure,
                    **extra_kwargs,
                )
                print("Submitted workflow run to prefect cloud.")

                # Attached the flow_run_id to our source so we know how many
                # associated jobs are running.
                source_db.prefect_flow_run_ids.append(flow_run_id)
                source_db.save()

                # update the source on the calculation
                # TODO: use the flow run id from above to grab the calc
                calculation = self.calculation_datatable.objects.get(
                    prefect_flow_run_id=flow_run_id
                )
                calculation.source = f"{source.__class__.__name__}"
                calculation.source_id = parent_ids
                calculation.save()

    def _make_new_structure(self, source, max_attempts=100):

        # check if we have a transformation or a creator
        # OPTIMIZE: is there a faster way? check subclass maybe?
        if "transformation" in str(type(source)):
            is_transformation = True
        elif "creator" in str(type(source)):
            is_transformation = False
        else:
            raise Exception(
                "Make sure your steady-state sources are either creators or transformations!"
                f" {source.__class__.__name__} does not meet this requirement."
            )

        # transformations require that we have completed structures in the
        # database. We want to wait until there's a set amount before
        # we start mutating the best. We check that here.
        if is_transformation and (
            self.search_datatable.individuals_completed.count() < self.nfirst_generation
        ):
            print(
                "Search isn't ready for transformations yet."
                f" Skipping {source.__class__.__name__}."
            )
            return False, False

        # Until we get a new valid structure (or run out of attempts), keep trying
        # with our given source. Assume we don't have a valid structure until
        # proven otherwise
        print(f"Attempting to create a structure with {source.__class__.__name__}")
        new_structure = False
        attempt = 0
        while not new_structure and attempt <= max_attempts:
            # add an attempt
            attempt += 1
            if is_transformation:
                # grab parent structures using the selection method
                parent_ids, parent_structures = self._select_parents(
                    nselect=source.ninput
                )
                # make a new structure
                new_structure = source.apply_transformation(parent_structures)
            elif not is_transformation:  # if it's a creator
                parent_ids = None
                # make a new structure
                new_structure = source.create_structure()

            # check to see if the structure is new and unique so that we don't
            # have any repeat calculations.
            if new_structure:
                # this will return false if the structure is NOT unique
                if not self.fingerprint_validator.check_structure(new_structure):
                    # if it is not unique, we can throw away the structure and
                    # try the loop again.
                    print("Generated structure is not unique. Trying again.")
                    new_structure = None

        # see if we got a structure or if we hit the max attempts and there's
        # a serious problem!
        if not new_structure:
            print(
                "Failed to create a structure! Consider changing your settings or"
                " contact our team for help."
            )
            return False, False

        # Otherwise we were successful
        print("Creation Successful.")

        # return the structure and its parents
        return parent_ids, new_structure

    def _select_parents(self, nselect):

        # our selectors just require a dataframe where we specify the fitness
        # column. So we query our individuals database to give this as an input.
        individuals_df = self.search_datatable.individuals_completed.order_by(
            "energy_per_atom"
        )[:200].to_dataframe()
        # NOTE: I assume we'll never need more than the best 200 structures, which
        # may not be true in special cases.

        # From these individuals, select our parent structures
        parents_df = self.selector.select(
            nselect,
            individuals_df,
            "energy_per_atom",
        )

        # grab the id column of the parents and convert it to a list
        parent_ids = parents_df.id.values.tolist()

        # Now lets grab these structures from our database and convert them
        # to a list of pymatgen structures.
        # We have to make separate queries for this instead of doing "id__in=parent_ids".
        # This is because (1) order may be important and (2) we may have duplicate
        # entries. For example, a hereditary mutation can request parent ids of [123,123]
        # in which case we want to give the same input structure twice!
        parent_structures = [
            self.search_datatable.individuals_completed.only("structure_string")
            .get(id=parent_id)
            .to_toolkit()
            for parent_id in parent_ids
        ]

        # When there's only one structure selected we return the structure and
        # id independently -- not within a list
        if nselect == 1:
            parent_ids = parent_ids[0]
            parent_structures = parent_structures[0]

        # for record keeping, we also want to return the ids for each structure
        return parent_ids, parent_structures

    def _init_common_class(self, class_str):

        # CREATORS
        if class_str in [
            "RandomSymStructure",
            "PyXtalStructure",
        ]:
            creator_class = getattr(creation_module, class_str)
            return creator_class(self.composition)

        # TRANSFORMATIONS
        elif class_str in [
            "from_ase.Heredity",
            "from_ase.SoftMutation",
            "from_ase.MirrorMutation",
            "from_ase.LatticeStrain",
            "from_ase.RotationalMutation",
            "from_ase.AtomicPermutation",
            "from_ase.CoordinatePerturbation",
        ]:
            # all start with "from_ase" so I assume that import for now
            ase_class_str = class_str.split(".")[-1]
            mutation_class = getattr(transform_module, ase_class_str)
            return mutation_class(self.composition)

        # !!! There aren't any common transformations that don't accept composition
        # as an input, but I expect this to change in the future.
        elif class_str in []:
            mutation_class = getattr(transform_module, class_str)
            return mutation_class()

        # These are commonly used single-shot sources
        elif class_str == "prototypes_aflow":
            pass  # TODO
        elif class_str == "substitution":
            pass  # TODO
        elif class_str == "third_party_structures":
            pass  # TODO
        elif class_str == "third_party_substitution":
            pass  # TODO

        else:
            raise Exception(
                f"{class_str} is not recognized as a common input. Make sure you"
                "don't have any typos, and if you are using a custom class, provide"
                "your input as an object."
            )



# -*- coding: utf-8 -*-

import os
import time
import datetime
import subprocess

import numpy
from pymatgen.io.vasp.outputs import Outcar

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.error_handlers import Unconverged


class Walltime(ErrorHandler):
    """
    Checks if a run is nearing the walltime of a SLURM job that it's
    running in. If so, VASP can be stopped by writing a STOPCAR with
    LSTOP or LABORT = True.

    Walltime is detected automatically based on environment variables set by
    SLURM, so if no variable is detected, this error handler has no effect.

    Support for PBS and other queue systems is not yet supported.

    For a similar implementation, see the `WallTimeHandler` in custodian located
    [here](https://github.com/materialsproject/custodian/blob/048adec8c965c1a01d69ac8294f039638941df74/custodian/vasp/handlers.py#L1521-L1629)
    """

    is_monitor = True
    has_custom_termination = True

    def __init__(
        self,
        wall_time: float = None,
        buffer_time: float = 900,  # 15 min
        electronic_step_stop: bool = False,
    ):
        """
        Initializes the handler with a buffer time.

        #### Parameters

        - `wall_time`:
            Total walltime for the calculation (in seconds). If this is None,
            and the job is running on a job-queue system, the handler will
            attempt to determine the remaining time using environment variables.
            If the wall time cannot be determined or is not set, this error
            handler will have no effect.

        - `buffer_time`:
            The minimum amount of time remaining to trigger this error handler.
            This is important because this error handler is only called
            periodically (e.g. once every 5 min). Therefore, if less than this
            buffer time remains, the error might not be caught because there
            won't be another check. When setting this, the S3Task's
            monitor_freq, polling_timestep, and time it takes to run all the
            other error handlers.

        - `electronic_step_stop`:
            Whether to check for electronic steps instead of ionic steps
            (e.g. for static runs on large systems or static HSE runs, ...).
            Be careful that results such as density or wavefunctions might
            not be converged at the electronic level. Should be used with
            LWAVE = .True. to be useful. If this is True, the STOPCAR is
            written with LABORT = .TRUE. instead of LSTOP = .TRUE.
        """
        self.wall_time = wall_time
        self.buffer_time = buffer_time
        self.electronic_step_stop = electronic_step_stop

    def check(self, directory: str) -> bool:
        """
        Check for error.
        """

        remaining_time = self._get_remaining_time(directory)

        # None means we ignore this error handler
        if remaining_time == None:
            return False

        # If the remaining time is less than our buffer time or time for 2
        # steps, then we also want to begin shutdown out of caution.
        time_per_step = self._get_max_step_time(directory)

        if (remaining_time < self.buffer_time) or (remaining_time < time_per_step * 2):
            # make sure the calculation didn't "finish at the buzzer"
            is_finished = self._check_if_finished(directory)
            if is_finished:
                print(
                    "BUZZER BEATER!!! This job finished right when it was "
                    "about to hit the walltime!"
                )
                return False
            # BUG: there may be a rare race condition here where the job completes
            # in the time it takes to indicate this check

            return True

        # Otherwise we don't have this error.
        return False

    def terminate_job(self, directory: str, **kwargs) -> bool:
        """
        When a walltime is about to be hit, we want to tell VASP to end
        naturally by creating a STOPCAR file. We do not want to allow and
        new VASP attempts, so we return a value of False for allow_retry
        """

        stopcar_filename = os.path.join(directory, "STOPCAR")
        stopcar_content = (
            "LSTOP = .TRUE." if not self.electronic_step_stop else "LABORT = .TRUE."
        )

        with open(stopcar_filename, "w") as stopcar:
            stopcar.write(stopcar_content)

        return False

    def correct(self):
        raise Exception("Stopped job due to Walltime limit.")
        # return "Added STOPCAR file to end job"  # ... and resubmitted

    def _get_remaining_time(self, directory: str) -> float:
        """
        Detects the amount of time remaining for a job by looking for class
        settings (e.g. wall_time), environment variables, or looking at file
        timestamps.

        Returns `None` if unable to detect the remaining time.
        """

        # A explicitly given wall time takes priority
        if self.wall_time:
            # check the remaining time by looking at the timestamp of the
            # simmate_metadata.yaml.
            # If there is no simmate_metadata.yaml (meaning the S3task
            # was ran outside of a Workflow), then we say that wall_time cannot
            # be determined. (checking files like INCAR won't work bc of other
            # handlers updating them)
            filename = os.path.join(directory, "simmate_metadata.yaml")
            if not os.path.exists(filename):
                print("Unable to detect the time remaining. Ignoring Timeout.")
                return
                # !!! What if I looked at the creation time of the current directory?

            time_since_creation = time.time() - os.path.getmtime(filename)

            remaining_time = self.wall_time - time_since_creation

            return remaining_time

        # Next we see if we are in a SLURM environment
        elif "SLURM_JOBID" in os.environ:

            output = subprocess.run(
                "squeue -h -O TimeLeft -j $SLURM_JOBID",
                shell=True,
                capture_output=True,
                text=True,
            ).stdout.strip()
            # parse the output into a time
            if output == "INVALID":
                print(
                    "WARNING: SLURM node improperly configured. "
                    "Cannot detect TimeLeft"
                )
                return
            elif output == "UNLIMITED":
                return
            else:
                # OPTIMIZE: this section could benefit from regex.

                # Converts a string like "199-21:58:12" to ['199', '21', '58', '12']
                try:
                    output = [int(i) for i in output.replace("-", ":").split(":")]
                except:
                    print(
                        "Failed to parse SLURM output. Please report this to the"
                        f"Simmate team. Output was {output}"
                    )
                    return
                # this funky format is because we don't know how long our list
                # will be.
                time_inputs = {}
                if len(output) == 4:
                    time_inputs["days"] = output[-4]
                if len(output) >= 3:
                    time_inputs["hours"] = output[-3]
                if len(output) >= 2:
                    time_inputs["minutes"] = output[-2]
                if len(output) >= 1:
                    time_inputs["seconds"] = output[-1]

                remaining = datetime.timedelta(**time_inputs).total_seconds()
                return remaining

        # TODO:
        # For PBS, I could use PBS_WALLTIME but I don't have a PBS system to
        # implement and test this out.

        # If nothing was triggered above, we ignore this handler.
        else:
            return None

    def _get_max_step_time(self, directory: str) -> float:
        """
        Looks at completed ionic/electronic steps to see the maximum time
        for each new step. Whether ion or electronic step time is returned
        depends on the electronic_step_stop setting.
        """

        outcar_filename = os.path.join(directory, "OUTCAR")
        outcar = Outcar(outcar_filename)

        # NOTE: this code is copied from the custodian error handler and
        # I choose not to mess with the regex. Parsing of time steps could
        # be moved to the Outcar class though.

        # Determine max time per ionic step.
        if not self.electronic_step_stop:

            outcar.read_pattern(
                {"timings": r"LOOP\+.+real time(.+)"}, postprocess=float
            )
            time_per_step = (
                numpy.max(outcar.data.get("timings"))
                if outcar.data.get("timings", [])
                else 0
            )

        # Determine max time per electronic step.
        else:

            outcar.read_pattern({"timings": "LOOP:.+real time(.+)"}, postprocess=float)
            time_per_step = (
                numpy.max(outcar.data.get("timings"))
                if outcar.data.get("timings", [])
                else 0
            )

        return time_per_step

    def _check_if_finished(self, directory: str):
        # We can invert the unconverged check in order to see if the calc is done.
        return not Unconverged().check(directory)
        # BUG: Will this work properly for MD simulations? Should I check the
        # number of steps completed?



# -*- coding: utf-8 -*-

"""
Workflows for predicting oxidation states within a material
"""

# TODO: when I add more calculators, I can do something like this...
# if "simmate.calculators.vasp" in installed_apps:
from simmate.calculators.vasp.workflows.population_analysis.all import (
    PopulationAnalysis__Vasp__BaderMatproj,
    PopulationAnalysis__Vasp__ElfMatproj,
    PopulationAnalysis__Vasp__BadelfMatproj,
)



# -*- coding: utf-8 -*-

from simmate.utilities import get_doc_from_readme

__doc__ = get_doc_from_readme(__file__)

# OPTIMIZE: These imports are particularly slow because we are configuring
# django in the code below this. These two imports don't depend on the database
# module, so it may be worth isolating them for faster imports.
from .error_handler import ErrorHandler
from .supervised_staged_shell_task import S3Task
from .worker import Worker

# All imports below this point depend on the simmate.database module and therefore
# need a database connection. This line sets up the database connection so
# that models can be imported.
from simmate.database import connect

from .workflow import Workflow, task



# -*- coding: utf-8 -*-

"""
This converter simply imports a pymatgen class. With this you can quickly
convert between Simmate.toolkit and CIF files.

This class is built-in to the base toolkit class, so you typically can read/write
with CIFs without ever loading this module directly.

Recommended use:
``` python

from simmate.toolkit import Structure

# read a cif file
initial_structure = Structure.from_file("example.cif")

# write a cif file
initial_structure.to("cif", "my_new_file.cif")

```

For advanced use, you can interact with this class directly:

``` python
from simmate.file_converters.structure.cif import CifParser

# load from file
cif_data = CifParser("example.cif")

# then access underlying data
cif_data.get_structures()
```
"""


from pymatgen.io.cif import CifParser, CifWriter

__all__ = ["CifParser", "CifWriter"]



# -*- coding: utf-8 -*-

from simmate.workflow_engine import Workflow
from simmate.calculators.vasp.tasks.density_of_states import (
    MatprojDensityOfStates as MatprojDensityOfStatesTask,
)
from simmate.calculators.vasp.database.density_of_states import (
    MatprojDensityOfStates as MatprojDensityOfStatesResults,
)


class ElectronicStructure__Vasp__MatprojDensityOfStates(Workflow):
    s3task = MatprojDensityOfStatesTask
    database_table = MatprojDensityOfStatesResults
    description_doc_short = "uses Materials Project settings (PBE)"



# -*- coding: utf-8 -*-

import os

import pytest

from simmate.command_line.database import database


@pytest.fixture  # BUG: is this test actually running...?
def test_database_reset(django_db_blocker, command_line_runner):

    # django-pytest does not let us to test the initial setup/configuration of the
    # database, so we need to do it within this context.
    #   https://pytest-django.readthedocs.io/en/latest/database.html#django-db-blocker
    with django_db_blocker.unblock():

        # reset the database
        result = command_line_runner.invoke(
            database, ["reset", "--confirm-delete", "--use-prebuilt=False"]
        )
        assert result.exit_code == 0

        # update the database
        result = command_line_runner.invoke(database, ["update"])
        assert result.exit_code == 0

        # OPTIMIZE: Is there a way to test this without downloading the large archive?
        # reset the database using a cloud archive
        # result = command_line_runner.invoke(
        #     database, ["reset", "--confirm-delete", "--use_prebuilt=True"]
        # )
        # assert result.exit_code == 0


@pytest.mark.django_db
def test_database_dump_and_load(command_line_runner):

    # dump the database to json
    result = command_line_runner.invoke(database, ["dumpdata"])
    assert result.exit_code == 0

    # load the database to json
    result = command_line_runner.invoke(database, ["loaddata"])
    assert result.exit_code == 0

    # delete the dump file
    os.remove("database_dump.json")



# -*- coding: utf-8 -*-

"""
This defines the base "simmate" command that all other commands stem from.
"""

import click

from simmate.command_line.database import database
from simmate.command_line.workflows import workflows
from simmate.command_line.workflow_engine import workflow_engine
from simmate.command_line.run_server import run_server
from simmate.command_line.start_project import start_project


@click.group()
def simmate():
    """
    This is the base command that all others stem from.

    If you are a beginner to the command line, be sure to start with our tutorials:
    https://github.com/jacksund/simmate/tree/main/tutorials

    Below you will see a list of sub-commands to try. For example, you can run `simmate
    database --help` to learn more about it.
    """
    # When we call the command "simmate" this is where we start, and it then
    # looks for all other functions that have the decorator "@simmate.command()"
    # to decide what to do from there.
    pass


# All commands are organized into other files, so we need to manually register
# them to our base "simmate" command here.
simmate.add_command(database)
simmate.add_command(workflows)
simmate.add_command(workflow_engine)
simmate.add_command(run_server)
simmate.add_command(start_project)

# explicitly list functions so that pdoc doesn't skip them
__all__ = ["simmate"]



# -*- coding: utf-8 -*-

import os
import json

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class Posmap(ErrorHandler):

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = ["POSMAP"]

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # load the error-count file if it exists
        error_count_filename = os.path.join(directory, "simmate_error_counts.json")
        if os.path.exists(error_count_filename):
            with open(error_count_filename) as error_count_file:
                error_counts = json.load(error_count_file)
        # otherwise we are starting with an empty dictionary
        else:
            error_counts = {}

        # The fix is based on the number of times we've already tried to
        # fix this. So let's make sure it's in our error_count dictionary.
        # If it isn't there yet, set the count to 0 and we'll update it below.
        error_counts["posmap"] = error_counts.get("posmap", 0) + 1

        # VASP advises to decrease or increase SYMPREC by an order of magnitude
        # the default SYMPREC value is 1e-5
        if error_counts["posmap"] == 0:
            # first, reduce by 10x
            current_symprec = incar.get("SYMPREC", 1e-5)
            new_symprec = current_symprec / 10
            incar["SYMPREC"] = new_symprec
            correction = f"switched SYMPREC from {current_symprec} to {new_symprec}"

        elif error_counts["posmap"] == 1:
            # next, increase by 100x (10x the original because we descreased
            # by 10x in the first try.)
            current_symprec = incar.get("SYMPREC", 1e-5)
            new_symprec = current_symprec * 100
            incar["SYMPREC"] = new_symprec
            correction = f"switched SYMPREC from {current_symprec} to {new_symprec}"

        # increase our attempt count
        error_counts["posmap"] += 1

        # rewrite the new error count file
        with open(error_count_filename, "w") as file:
            json.dump(error_counts, file)

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return correction



# -*- coding: utf-8 -*-

from inspect import signature


class TriggeredAction:
    def __init__(
        self,
    ):
        # add any unchanging variables or settings here
        pass

    def check(self, search):
        # This method should take in one argument:
        # search = this is the main Search object seen in search.py
        # The Trigger can run any analysis on the Search object WITHOUT
        # making any changes to it. Then you should return True if the
        # trigger condition(s) has been met and False if the calculation should
        # continue without setting off the trigger
        pass

    def action(self, search):
        # This method should take in one argument:
        # search = this is the main Search object seen in search.py
        # This is where to apply some change to the search. It can be whatever you'd like!
        # The most common actions (such as search.new_sample) will be accessed via
        # the search class and not custom code here.
        pass

    @classmethod
    def from_composition(cls, composition=None, **kwargs):

        # note that trigger_options is a dict of custom inputs to use on the class trigger_class

        import pymatdisc.engine.triggers as trigger_module

        # if the trigger class is a string, then assume we want to import from the trigger_module
        if type(cls) == str:
            trigger_class = getattr(trigger_module, trigger_class)
        # otherwise, the user is trying to use their own module/class and we already have that set

        # now that we have the class, we want to initiate it with the settings provided
        # we also need to see if composition is a required input, which we don't require the user to specify for convenience
        trigger_class_parameters = signature(trigger_class).parameters

        #!!! in the future, should I just require that all mutators take composition (or **kwargs) as an input even if they don't need it?
        # now we can init with the dictionary of options depending on if composition is needed or not
        if "composition" in trigger_class_parameters:
            trigger_object = trigger_class(composition, **trigger_options)
        else:
            trigger_object = trigger_class(**trigger_options)

        # we now have the final trigger object instance and can return it
        return trigger_object


#!!! In the future, should I reduce triggers down to lower order objects?
#!!! One example of this would be reducing a Trigger to a Check/Validation and an Action

# -----------------------------------------------------------------------------


class InitStructures:

    # Super simple check that passes True if no structures have been created yet.
    # If that is the case, then it creates new structures using the input probilities for each generator.
    # This is used to start the entire search.

    def __init__(self, n_initial_structures):

        # Number of initial structures to create
        self.n_initial_structures = n_initial_structures

    def check(self, search):  #!!! can I move 'search' to the init?

        if not search.structures:  # same as len(structures) == 0
            return True
        else:
            return False

    def action(self, search):

        print("Making new structures...")

        # we want n total structures so we are going to loop this number of times
        for n in range(self.n_initial_structures):
            # make a new sample
            search.new_sample(creators_only=True)


# -----------------------------------------------------------------------------


class AddStructures:
    def __init__(self, n_pending_limit, n_add_structures):

        self.n_pending_limit = n_pending_limit

        # Number of structures to create if the pending_limit is hit
        self.n_add_structures = n_add_structures

    def check(self, search):  #!!! can I move 'search' to the init?

        # if no structures have completed yet, we don't want to add any new structure with this trigger
        if search.njobs_completed == 0:
            return False

        # See if the number of jobs pending has dropped below the limit
        if search.njobs_pending <= self.n_pending_limit:
            return True
        else:
            return False

    def action(self, search):

        print("Making new structures...")

        # we want n total structures so we are going to loop this number of times
        for n in range(self.n_add_structures):
            # make a new sample
            search.new_sample()


# -----------------------------------------------------------------------------



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import table_column, Structure, Thermodynamics


class AflowStructure(Structure, Thermodynamics):
    """
    Crystal structures from the [AFLOW](http://aflowlib.org/) database.

    Currently, this table only stores strucure and thermodynamic information,
    but the AFLOW has much more data available via their
    [REST API](http://aflowlib.duke.edu/aflowwiki/doku.php?id=documentation:start)
    and website.
    """

    class Meta:
        app_label = "third_parties"

    base_info = ["id", "structure_string", "energy"]
    source = "AFLOW"
    source_doi = "https://doi.org/10.1016/j.commatsci.2012.02.005"

    id = table_column.CharField(max_length=25, primary_key=True)
    """
    The id used to represent the structure (ex: "aflow-12345")
    """

    @property
    def external_link(self) -> str:
        """
        URL to this structure in the AFLOW website.
        """
        # Links to the AFLOW dashboard for this structure. An example is...
        #   http://aflow.org/material/?id=aflow:ffea9279661c929f
        # !!! I could also consider an alternative link which points to an interactive
        # REST endpoint. The API is also sporatic for these, An example one is...
        #   aflowlib.duke.edu/AFLOWDATA/ICSD_WEB/FCC/Dy1Mn2_ICSD_602151
        id_formatted = self.id.replace("-", ":")
        return f"http://aflow.org/material/?id={id_formatted}"



# -*- coding: utf-8 -*-

import pytest

from simmate.workflows.utilities import (
    get_list_of_all_workflows,
    get_workflow,
)


@pytest.mark.vasp
@pytest.mark.prefect_db
@pytest.mark.django_db
def test_all_workflow_runs(tmpdir, sample_structures):

    # For testing, look at the NaCl rocksalt primitive structure
    structure = sample_structures["NaCl_mp-22862_primitive"]

    with tmpdir.as_cwd():

        successful_flows = []

        # TEST CUSTOMIZED FLOW
        workflow_name = "customized.vasp.user-config"
        workflow = get_workflow(workflow_name)
        state = workflow.run(
            workflow_base="relaxation.vasp.quality00",
            input_parameters={
                "structure": structure,
                "command": "mpirun -n 12 vasp_std > vasp.out",
            },
            updated_settings={
                "incar": {"NPAR": 1, "ENCUT": 600},
            },
        )
        if state.is_completed():
            successful_flows.append(workflow_name)

        # TEST S3Workflows
        s3_flows = [
            "electronic-structure.vasp.matproj-full",
            "population-analysis.vasp.bader-matproj",
            "population-analysis.vasp.elf-matproj",
            "relaxation.vasp.matproj",
            "relaxation.vasp.mit",
            "relaxation.vasp.neb-endpoint",
            "relaxation.vasp.quality00",
            "relaxation.vasp.quality01",
            "relaxation.vasp.quality02",
            "relaxation.vasp.quality03",
            "relaxation.vasp.staged",
            "relaxation.vasp.quality04",
            "static-energy.vasp.matproj",
            "static-energy.vasp.mit",
            "static-energy.vasp.neb-endpoint",
            "static-energy.vasp.quality04",
        ]
        for workflow_name in s3_flows:

            workflow = get_workflow(workflow_name)
            state = workflow.run(
                structure=structure,
                command="mpirun -n 12 vasp_std > vasp.out",
            )
            if state.is_completed():
                successful_flows.append(workflow_name)

        # TEST MD FLOW
        workflow_name = "dynamics.vasp.mit"
        workflow = get_workflow(workflow_name)
        state = workflow.run(
            structure=structure,
            command="mpirun -n 12 vasp_std > vasp.out",
            nsteps=100,
            temperature_start=400,
            temperature_end=400,
        )
        if state.is_completed():
            successful_flows.append(workflow_name)

        # TEST NEB FLOWS
        # For testing, look at I- diffusion in Y2CF2
        # structure = sample_structures["Y2CI2_mp-1206803_primitive"]
        # workflow_name = "diffusion.vasp.neb-all-paths"
        # workflow = get_workflow(workflow_name)
        # state = workflow.run(
        #     structure=structure,
        #     migrating_specie="I",
        #     command="mpirun -n 15 vasp_std > vasp.out",
        #     directory=str(tmpdir),
        # )
        # state.result()
        # if state.is_completed():
        #     successful_flows.append(workflow_name)

    # check which flows either (1) failed or (2) weren't tested
    all_flows = get_list_of_all_workflows()
    missing_failed_flows = list(set(all_flows) - set(successful_flows))
    missing_failed_flows.sort()

    assert missing_failed_flows == [
        "diffusion.vasp.neb-all-paths",
        "diffusion.vasp.neb-from-endpoints",
        "diffusion.vasp.neb-from-images",
        "diffusion.vasp.neb-single-path",
        "population-analysis.vasp.badelf-matproj",
    ]



# -*- coding: utf-8 -*-

from simmate.workflow_engine import Workflow
from simmate.calculators.vasp.tasks.static_energy import (
    MatprojStaticEnergy as MPStaticEnergyTask,
)
from simmate.calculators.vasp.database.energy import (
    MatprojStaticEnergy as MPStaticEnergyResults,
)


class StaticEnergy__Vasp__Matproj(Workflow):
    s3task = MPStaticEnergyTask
    database_table = MPStaticEnergyResults
    description_doc_short = "uses Materials Project settings"



# -*- coding: utf-8 -*-

import os

import pytest

from simmate.conftest import copy_test_files
from simmate.calculators.vasp.inputs import Incar
from simmate.calculators.vasp.error_handlers import LargeSigma


def test_large_sigma(tmpdir):
    copy_test_files(
        tmpdir,
        test_directory=__file__,
        test_folder="large_sigma",
    )

    # we reference the files several spots below so we grab its path up front
    incar_filename = os.path.join(tmpdir, "INCAR")
    outcar_filename = os.path.join(tmpdir, "OUTCAR")

    # init class with default settings
    error_handler = LargeSigma()

    # Confirm an error IS found
    assert error_handler.check(tmpdir) == True

    # Make first attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "reduced SIGMA from 0.1 to 0.04000000000000001"
    assert Incar.from_file(incar_filename)["SIGMA"] == 0.04000000000000001

    # Make 2nd attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "switched KSPACING from 0.5 to 0.4"
    assert Incar.from_file(incar_filename)["KSPACING"] == 0.4

    # Make final attempt at fixing the error, which raises an error
    incar = Incar.from_file(incar_filename)
    incar["KSPACING"] = 0.2
    incar.to_file(incar_filename)
    with pytest.raises(Exception):
        fix = error_handler.correct(tmpdir)

    # Confirm an error IS NOT found when no outcar exists
    os.remove(outcar_filename)
    assert error_handler.check(tmpdir) == False

    # Confirm an error IS NOT found when ISMEAR > 0
    incar = Incar.from_file(incar_filename)
    incar["ISMEAR"] = -1
    incar.to_file(incar_filename)
    assert error_handler.check(tmpdir) == False



# -*- coding: utf-8 -*-

import os

import pytest

from simmate.conftest import copy_test_files
from simmate.calculators.vasp.inputs import Incar
from simmate.calculators.vasp.error_handlers import NonConverging


def test_nonconverging(tmpdir):
    copy_test_files(
        tmpdir,
        test_directory=__file__,
        test_folder="nonconverging",
    )

    # we reference the files several spots below so we grab its path up front
    incar_filename = os.path.join(tmpdir, "INCAR")
    oszicar_filename = os.path.join(tmpdir, "OSZICAR")

    # Confirm an error IS NOT found when there aren't enough ionic steps
    error_handler = NonConverging()  # default is 10
    assert error_handler.check(tmpdir) == False

    # We limit this to 2 steps to ensure we get the error
    error_handler = NonConverging(min_ionic_steps=2)
    assert error_handler.check(tmpdir) == True

    # Make first attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "switched ALGO from VeryFast to Fast"
    assert Incar.from_file(incar_filename)["ALGO"] == "Fast"

    # Make second attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "switched ALGO from Fast to Normal"
    assert Incar.from_file(incar_filename)["ALGO"] == "Normal"

    # Make third attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "switched ALGO from Normal to All"
    assert Incar.from_file(incar_filename)["ALGO"] == "All"

    # Make fourth attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "switched linear mixing via {'AMIX': 0.1, 'BMIX': 0.01, 'ICHARG': 2}"
    incar = Incar.from_file(incar_filename)
    assert incar["AMIX"] == 0.1
    assert incar["BMIX"] == 0.01
    assert incar["ICHARG"] == 2

    # Make fifth attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "switched linear mixing via {'AMIN': 0.01, 'BMIX': 3.0, 'ICHARG': 2}"
    incar = Incar.from_file(incar_filename)
    assert incar["AMIN"] == 0.01
    assert incar["BMIX"] == 3.0
    assert incar["ICHARG"] == 2

    # Make final attempt at fixing the error, which raises an error
    with pytest.raises(Exception):
        fix = error_handler.correct(tmpdir)

    # Make sure an error is not raised when an electronic step converges
    incar = Incar.from_file(incar_filename)
    incar["NELM"] = 2
    incar.to_file(incar_filename)
    assert error_handler.check(tmpdir) == False

    # Make sure an error is not raised when no OSZICAR is present
    os.remove(oszicar_filename)
    assert error_handler.check(tmpdir) == False



# -*- coding: utf-8 -*-

from simmate.workflow_engine import Workflow
from simmate.calculators.vasp.tasks.static_energy import (
    Quality04Energy as Quality04EnergyTask,
)
from simmate.calculators.vasp.database.energy import (
    Quality04StaticEnergy as Quality04StaticEnergyResults,
)


class StaticEnergy__Vasp__Quality04(Workflow):
    s3task = Quality04EnergyTask
    database_table = Quality04StaticEnergyResults
    description_doc_short = (
        "low-quality settings meant for highly unreasonable structures"
    )



# -*- coding: utf-8 -*-

from django.shortcuts import render
from django.db.models import F
from django.contrib.auth.decorators import login_required

from simmate.website.third_parties.forms import ChemicalSystemForm
from simmate.database.third_parties import (
    AflowStructure,
    CodStructure,
    JarvisStructure,
    MatprojStructure,
    OqmdStructure,
)


@login_required
def profile(request):
    # !!! For future reference, you can grab user-associated data via...
    # data = request.user.relateddata.all()

    context = {}
    template = "account/profile.html"
    return render(request, template, context)


def loginstatus(request):
    context = {}
    template = "account/loginstatus.html"
    return render(request, template, context)


def home(request):

    # The home page is also an html "form" because users submit queries from
    # here. So we need to handle form submissions properly.

    # first check if the webpage is accessed via a POST method
    if request.method == "POST":
        # if it is, that means a user is trying to submit a query
        # let's grab the data and validate it before running the query
        form = ChemicalSystemForm(request.POST)
        # see if all of the data is valid
        if form.is_valid():

            # grab the cleaned data from the form
            cleaned_data = form.cleaned_data

            # check which databases the user wants to search and collect the
            # proper models into a list that we query with below
            databases_to_search = []
            for database, database_model in (
                ("aflow", AflowStructure),
                ("cod", CodStructure),
                ("jarvis", JarvisStructure),
                ("materials_project", MatprojStructure),
                ("oqmd", OqmdStructure),
            ):
                # if the user requested this database, the value will be true
                if cleaned_data[database]:
                    databases_to_search.append(database_model)

            # Note, this value is converted to a list of systems once cleaned.
            chemical_systems = cleaned_data["chemical_system"]

            # now go through each database, search for the requested system
            # and then pool them all together. We limit each database to 50
            # results. Note the += below means we compile into a single list
            structures = []
            nstructures_possible = 0
            for database_model in databases_to_search:
                # Now we can make the query! We also dont want to load the
                # structure json -- so that everything runs faster.
                search_results = database_model.objects.filter(
                    chemical_system__in=chemical_systems
                ).defer("structure_string")

                # if the database provides the hull energy, we want to sort
                # the structures by that (putting highest priority on stable ones)
                if hasattr(database_model, "energy_above_hull"):
                    # if there isn't a hull energy value, place these last
                    search_results = search_results.order_by(
                        F("energy_above_hull").asc(nulls_last=True)
                    )

                # now add the search results to the output
                # for performance, we limit each database to 50 structures
                structures += search_results.all()[:50]

                # We also tell the user how many results are possible if there
                # wasn't any limit on the structures returned
                nstructures_possible += database_model.objects.filter(
                    chemical_system__in=chemical_systems,
                ).count()

    # if the page is grabbed via a 'GET' method, send an empty form
    else:
        # otherwise we are giving an empty form and no result structures
        form = ChemicalSystemForm()
        structures = None
        nstructures_possible = None

    # now let's put the data and template together to send the user
    context = {
        "chemical_system_form": form,
        "structures": structures,
        "nstructures_possible": nstructures_possible,
    }
    template = "home/home.html"
    return render(request, template, context)


def extras(request):
    context = {}
    template = "core_components/extras.html"
    return render(request, template, context)


def contact(request):
    context = {}
    template = "core_components/contact.html"
    return render(request, template, context)


def about(request):
    context = {}
    template = "core_components/about.html"
    return render(request, template, context)



# -*- coding: utf-8 -*-

from django_filters import rest_framework as filters

from simmate.website.core_components.filters import Calculation, Structure
from simmate.database.base_data_types import (
    BandStructure as BandStructureTable,
    BandStructureCalc as BandStructureCalcTable,
)


class BandStructure(filters.FilterSet):
    class Meta:
        model = BandStructureTable
        fields = dict(
            nbands=["range"],
            band_gap=["range"],
            band_gap_direct=["range"],
            is_gap_direct=["exact"],
            energy_fermi=["range"],
            conduction_band_minimum=["range"],
            valence_band_maximum=["range"],
            is_metal=["exact"],
        )


class BandStructureCalc(BandStructure, Calculation, Structure):
    class Meta:
        model = BandStructureCalcTable
        fields = {
            **Structure.get_fields(),
            **BandStructure.get_fields(),
            **Calculation.get_fields(),
        }



# -*- coding: utf-8 -*-

import os
import pytest

from simmate.conftest import make_dummy_files
from simmate.calculators.vasp.inputs import Incar, Potcar
from simmate.toolkit.diffusion import MigrationImages

from pymatgen.io.vasp.sets import (
    MPRelaxSet,
    MITRelaxSet,
    MPStaticSet,
    MITMDSet,
    MPMDSet,
    MPHSERelaxSet,
    MPHSEBSSet,
    MPNonSCFSet,
    MPMetalRelaxSet,
    MPNMRSet,
    MVLElasticSet,
    MVLGBSet,
    MITNEBSet,
    # MVLNPTMDSet,
)
from pymatgen.analysis.diffusion.neb.io import MVLCINEBSet, MVLCINEBEndPointSet

from simmate.calculators.vasp.tasks.static_energy import MatprojStaticEnergy
from simmate.calculators.vasp.tasks.relaxation import (
    MatprojRelaxation,
    MITRelaxation,
    MatprojHSERelaxation,
    MatprojMetalRelaxation,
    MatVirtualLabCINEBEndpointRelaxation,
)
from simmate.calculators.vasp.tasks.dynamics import (
    MITDynamics,
    MatprojDynamics,
    # MatVirtualLabNPTDynamics,
)
from simmate.calculators.vasp.tasks.band_structure import (
    MatprojHSEBandStructure,
    MatprojBandStructure,
)
from simmate.calculators.vasp.tasks.density_of_states import (
    MatprojHSEDensityOfStates,
    MatprojDensityOfStates,
)
from simmate.calculators.vasp.tasks.nuclear_magnetic_resonance import (
    MatprojNMRChemicalShifts,
    # MatprojNMRElectricFieldGradiant,
)
from simmate.calculators.vasp.tasks.elastic import MatVirtualLabElastic
from simmate.calculators.vasp.tasks.relaxation import (
    MatVirtualLabGrainBoundaryRelaxation,
    MatVirtualLabSlabRelaxation,
)
from simmate.calculators.vasp.tasks.nudged_elastic_band import (
    MITNudgedElasticBand,
    MatVirtualLabClimbingImageNudgedElasticBand,
)


MD_KWARGS = {
    "start_temp": 300,
    "end_temp": 1200,
    "nsteps": 10000,
}


@pytest.mark.pymatgen
@pytest.mark.parametrize(
    "simmate_task, pymatgen_set, pymatgen_kwargs",
    [
        (MatprojRelaxation, MPRelaxSet, {}),
        (MITRelaxation, MITRelaxSet, {}),
        (MITRelaxation, MITRelaxSet, {}),
        (MatprojStaticEnergy, MPStaticSet, {}),
        (MatprojHSERelaxation, MPHSERelaxSet, {}),
        (MatprojMetalRelaxation, MPMetalRelaxSet, {}),
        (MatVirtualLabCINEBEndpointRelaxation, MVLCINEBEndPointSet, {}),
        (MITDynamics, MITMDSet, MD_KWARGS),
        (MatprojDynamics, MPMDSet, MD_KWARGS),
        # (MatVirtualLabNPTDynamics, MVLNPTMDSet, MD_KWARGS),
        # requires POTCARs to be configured for pymatgen. need a workaround.
        (MatprojHSEBandStructure, MPHSEBSSet, {"mode": "line"}),
        (MatprojBandStructure, MPNonSCFSet, {"mode": "line"}),
        (MatprojDensityOfStates, MPNonSCFSet, {"mode": "uniform"}),
        (MatprojHSEDensityOfStates, MPHSEBSSet, {"mode": "uniform"}),
        (MatprojNMRChemicalShifts, MPNMRSet, {"mode": "cs"}),
        # (MatprojNMRElectricFieldGradiant, MPNMRSet, {"mode": "efg"}),
        # MPNMRSet is bugged at the moment: https://github.com/materialsproject/pymatgen/pull/2509
        (MatVirtualLabElastic, MVLElasticSet, {}),
        (MatVirtualLabGrainBoundaryRelaxation, MVLGBSet, {"slab_mode": False}),
        (MatVirtualLabSlabRelaxation, MVLGBSet, {"slab_mode": True}),
    ],
)
def test_pymatgen_input_sets(
    structure,
    tmpdir,
    mocker,
    simmate_task,
    pymatgen_set,
    pymatgen_kwargs,
):
    """
    Many of the presets implemented are ported directly from pymatgen, so these
    tests simply confirm that the inputs give the same INCAR results.
    """

    # set filenames
    incar_simmate_name = os.path.join(tmpdir, "INCAR")
    incar_pymatgen_name = os.path.join(tmpdir, "INCAR_pmg")

    # Because we won't have POTCARs accessible, we need to cover this function
    # call -- specifically have it pretend to make a file
    potcar_filename = os.path.join(tmpdir, "POTCAR")
    mocker.patch.object(
        Potcar,
        "to_file_from_type",
        return_value=make_dummy_files(potcar_filename),
    )

    # write both inputs
    simmate_task().setup(structure=structure, directory=tmpdir)
    pymatgen_set(structure, **pymatgen_kwargs).incar.write_file(incar_pymatgen_name)

    # load incar
    incar_simmate = Incar.from_file(incar_simmate_name)
    incar_pymatgen = Incar.from_file(incar_pymatgen_name)

    # compare
    diff = incar_simmate.compare_incars(incar_pymatgen)["Different"]
    # ignore kpoints input for now
    diff.pop("KSPACING", None)

    # in EDIFF__per_atom, there are sometimes rounding issues.
    # We say there is no difference if they are within 100% error. This may
    # seem like a large error, but it is meant to encompass differences like
    # {'EDIFF': (4e-06, 1e-05)} where pymatgen rounded significantly. There
    # really isn't an issue unless EDIFF is off by a factor of 10 (1000%)
    if "EDIFF" in diff and "EDIFF__per_atom" in simmate_task.incar:
        ediff_simmate, ediff_pymatgen = diff["EDIFF"]
        percent_diff = (ediff_pymatgen - ediff_simmate) / ediff_pymatgen
        if abs(percent_diff) < 1:
            diff.pop("EDIFF")

    # ensure incars are the same (there are no differences)
    assert diff == {}


@pytest.mark.pymatgen
@pytest.mark.parametrize(
    "simmate_task, pymatgen_set, pymatgen_kwargs",
    [
        (MITNudgedElasticBand, MITNEBSet, {}),
        (MatVirtualLabClimbingImageNudgedElasticBand, MVLCINEBSet, {}),
    ],
)
def test_pymatgen_input_sets_neb(
    structure,
    tmpdir,
    mocker,
    simmate_task,
    pymatgen_set,
    pymatgen_kwargs,
):
    """
    This is a copy/paste of the test method above, but this is for sets that
    accept multiple structures as inputs (such as NEB sets)
    """

    # set filenames
    incar_simmate_name = os.path.join(tmpdir, "INCAR")
    incar_pymatgen_name = os.path.join(tmpdir, "INCAR_pmg")

    # Because we won't have POTCARs accessible, we need to cover this function
    # call -- specifically have it pretend to make a file
    potcar_filename = os.path.join(tmpdir, "POTCAR")
    mocker.patch.object(
        Potcar,
        "to_file_from_type",
        return_value=make_dummy_files(potcar_filename),
    )

    # write both inputs. We just copy the same structure 3 times because we
    # don't really care what the POSCARs look like -- just the INCARs
    simmate_task().setup(
        migration_images=MigrationImages([structure] * 3),
        directory=tmpdir,
    )
    pymatgen_set([structure] * 3, **pymatgen_kwargs).incar.write_file(
        incar_pymatgen_name
    )

    # load incar
    incar_simmate = Incar.from_file(incar_simmate_name)
    incar_pymatgen = Incar.from_file(incar_pymatgen_name)

    # compare
    diff = incar_simmate.compare_incars(incar_pymatgen)["Different"]
    # ignore kpoints input for now
    diff.pop("KSPACING", None)

    # in EDIFF__per_atom, there are sometimes rounding issues.
    # We say there is no difference if they are within 100% error. This may
    # seem like a large error, but it is meant to encompass differences like
    # {'EDIFF': (4e-06, 1e-05)} where pymatgen rounded significantly. There
    # really isn't an issue unless EDIFF is off by a factor of 10 (1000%)
    if "EDIFF" in diff and "EDIFF__per_atom" in simmate_task.incar:
        ediff_simmate, ediff_pymatgen = diff["EDIFF"]
        percent_diff = (ediff_pymatgen - ediff_simmate) / ediff_pymatgen
        if abs(percent_diff) < 1:
            diff.pop("EDIFF")

    # ensure incars are the same (there are no differences)
    assert diff == {}



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import StaticEnergy


class MITStaticEnergy(StaticEnergy):
    pass


class MatprojStaticEnergy(StaticEnergy):
    pass


class NEBEndpointStaticEnergy(StaticEnergy):
    pass


class Quality04StaticEnergy(StaticEnergy):
    pass



# -*- coding: utf-8 -*-

from simmate.utilities import get_doc_from_readme

__doc__ = get_doc_from_readme(__file__)

from .incar import Incar
from .kpoints import Kpoints
from .potcar import Potcar
from .poscar import Poscar



# -*- coding: utf-8 -*-

from simmate.workflow_engine import Workflow
from simmate.calculators.vasp.tasks.relaxation import (
    MatVirtualLabCINEBEndpointRelaxation as NEBEndpointRelaxationTask,
)
from simmate.calculators.vasp.database.relaxation import (
    MatVirtualLabCINEBEndpointRelaxation as NEBEndpointRelaxationResults,
)


class Relaxation__Vasp__NebEndpoint(Workflow):
    s3task = NEBEndpointRelaxationTask
    database_table = NEBEndpointRelaxationResults
    description_doc_short = (
        "uses Materials Project settings and meant for endpoint supercells"
    )



# -*- coding: utf-8 -*-

"""

Currently I grab the first table for each spacegroup, but in many cases there 
are more options. For example, spacegroup 10 has "Unique Axis b" and "Unique 
Axis c" tables to choose from -- by default I grab the first. The same goes for
spacegroups like 166 where there is a "Hexagonal" and "Rhombohedral" option. 
If instead I wanted to specify which to choose, I would look instead the headers
on each page. The headers should line up with the tables. For example in
spacegroup 166, I can see:
    if "HEXAGONAL AXES" in header.text

"""


from selenium import webdriver
import pandas as pd
from tqdm import tqdm

# launch the webbrowser
driver = webdriver.Chrome(
    executable_path="/snap/bin/chromium.chromedriver"
)  # /snap/bin/chromium gives errors

asym_data = []
wy_data = []
for spacegroup in tqdm(range(1, 231)):

    # change the int spacegroup to a string in three digit format
    # for example, 15 will turn into 015 and 1 will turn into 001
    sg_format = (
        f"{spacegroup:03}"  # I don't understand f-string formatting yet but this works
    )

    # make our target url
    target_url = "https://it.iucr.org/Ab/ch7o1v0001/sgtable7o1o" + sg_format + "/"

    # load the webpage
    driver.get(target_url)

    ###### Grab the asymmetric unit data ######

    # In many cases there are multiple tables becuase there are multiple ways to draw a lattice
    # For example, spacegroup 10 has two options: "UNIQUE AXIS b" and "UNIQUE AXIS c"
    # I still don't fully understand why there are two in this case
    # this issue more clear when looking at spacegroup 166, which has two options: "HEXAGONAL AXES" and "RHOMBOHEDRAL AXES"
    # In all cases except for when it's hex vs. rhomb, we want the first table

    # first grab everything labeled with class="asymmetricunit"
    elements = driver.find_elements_by_class_name("asymmetricunit")
    # then limit these elements to those that possess the <table> tag
    tables = [element for element in elements if element.tag_name == "table"]
    # grab everything labeled with class="sgheader"
    elements = driver.find_elements_by_class_name("sgheader")
    # then limit these elements to those that possess the <div> tag
    headers = [element for element in elements if element.tag_name == "div"]
    # we can use the header to check for which cell type / unique axis table we want
    # but for now, we are just grabbing the first table
    target = 0
    # grab the header for our target table and make it into a list of strings
    header = headers[target].text.split("\n")
    # cell specification is only in some cases (i.e. a third row doesn't always exist)
    if len(header) > 2:
        cell_spec = header[2]
    else:
        cell_spec = None
    # the data we want is in the 2nd cell of the 1st table
    data = tables[target].find_elements_by_tag_name("td")[1].text
    # append to output list
    asym_data.append([spacegroup, data, cell_spec])

    ###### Grab the wyckoff site data ######

    # we are following the same rules as with the asymmetric data above (first table unless its hexagonal)
    # There are two table we are grabbing from here -- genpos as well as specpos

    # first grab everything labeled with class="genpos"
    elements = driver.find_elements_by_class_name("genpos")
    # then limit these elements to those that possess the <table> tag
    tables = [element for element in elements if element.tag_name == "table"]
    # grab everything labeled with class="sgheader"
    elements = driver.find_elements_by_class_name("sgheader")
    # then limit these elements to those that possess the <div> tag
    headers = [element for element in elements if element.tag_name == "div"]
    # we can use the header to check for which cell type / unique axis table we want
    # but for now, we are just grabbing the first table
    target = 0
    # grab the header for our target table and make it into a list of strings
    header = headers[target].text.split("\n")
    # grab all of the rows in the 1st table
    rows = tables[target].find_elements_by_tag_name("tr")
    # we only want rows with class="specpos"
    # and we are looking specifically at the text
    rows = [row.text for row in rows if row.get_attribute("class") == "specpos"]
    # the first row is the header so skip it
    for row in rows[1:]:
        # change characters in the string and then split it into a list
        row_content = row.replace("\n", " ").split(" ")
        # there is a bug for spacegroup 47, because its letter is alpha - which doesnt show up
        if spacegroup == 47:
            row_content.insert(1, "A")
        # first index is Multiplicity
        mult_conv = int(row_content[0])
        # second index is Wyckoffletter
        # third is Sitesymmetry -- which for general positions is always 1
        letter, sym = row_content[1:3]
        # fifth through seventh are the coords -- which for general is always (x,y,z)
        coords = row_content[4] + row_content[5] + row_content[6]
        coords = coords.replace("2x", "2*x")
        # Note the that the current Multiplicity indicated is for the convential cell
        # where we'd like to know the primitive cell's multiplicity
        # The letters marked below correspond to the following:
        # P primitive
        # I body centered
        # F face centered
        # A centered on A faces only
        # B centered on B faces only
        # C centered on C faces only
        # R rhombohedral
        # I grab this information from PyXtal's crystal.py file.
        if spacegroup in [
            22,
            42,
            43,
            69,
            70,
            196,
            202,
            203,
            209,
            210,
            216,
            219,
            225,
            226,
            227,
            228,
        ]:
            mult_prim = mult_conv / 4  # F
        elif spacegroup in [146, 148, 155, 160, 161, 166, 167]:
            mult_prim = mult_conv / 3  # R
        elif spacegroup in [
            5,
            8,
            9,
            12,
            15,
            20,
            21,
            23,
            24,
            35,
            36,
            37,
            38,
            39,
            40,
            41,
            44,
            45,
            46,
            63,
            64,
            65,
            66,
            67,
            68,
            71,
            72,
            73,
            74,
            79,
            80,
            82,
            87,
            88,
            97,
            98,
            107,
            108,
            109,
            110,
            119,
            120,
            121,
            122,
            139,
            140,
            141,
            142,
            197,
            199,
            204,
            206,
            211,
            214,
            217,
            220,
            229,
            230,
        ]:
            mult_prim = mult_conv / 2  # A, C, I
        else:
            mult_prim = mult_conv / 1  # P
        # append to output list
        wy_data.append([spacegroup, mult_conv, mult_prim, letter, sym, coords])

    # now grab everything labeled with class="specpos"
    elements = driver.find_elements_by_class_name("specpos")
    # then limit these elements to those that possess the <table> tag
    tables = [element for element in elements if element.tag_name == "table"]
    # if there are no tables, jump to the next spacegroup (this happens with spacegroup 1)
    if not tables:
        continue
    # grab everything labeled with class="sgheader"
    elements = driver.find_elements_by_class_name("sgheader")
    # then limit these elements to those that possess the <div> tag
    headers = [element for element in elements if element.tag_name == "div"]
    # we can use the header to check for which cell type / unique axis table we want
    # but for now, we are just grabbing the first table
    target = 0
    # grab the header for our target table and make it into a list of strings
    header = headers[target].text.split("\n")
    # grab all of the rows in the 2nd table
    rows = tables[target].find_elements_by_tag_name("tr")
    # we only want rows with class="specpos"
    # and we are looking specifically at the text
    rows = [row for row in rows if row.get_attribute("class") == "specpos"]
    # the first row is the header so skip it
    for row in rows[1:]:
        # grab the text; change characters in the string; and then split it into a list
        row_content = row.text.replace("\n", " ").split(" ")
        # first index is Multiplicity
        mult_conv = int(row_content[0])
        # second index is Wyckoffletter
        letter = row_content[1]
        # fourth and the ones that follow are Sitesymmetry
        # # third is empty and Sitesymmtry label ends at the next empty
        sym = ""
        i = 3
        while row_content[i] != "":
            sym += row_content[i]
            i += 1
        # in order to grab the coordinates, we need to look at all the entries inside of the row
        row_elements = row.find_elements_by_tag_name("td")
        # and we want the elements specifically with class='specposcoords'
        row_elements = [
            element
            for element in row_elements
            if element.get_attribute("class") == "specposcoords"
        ]
        # and the coords that we are interested in should be the first entry
        # I also want to remove the spaces in the coords
        coords = row_elements[0].text.replace(" ", "").replace("2x", "2*x")
        # Note the that the current Multiplicity indicated is for the convential cell
        # where we'd like to know the primitive cell's multiplicity
        # The letters marked below correspond to the following:
        # P primitive
        # I body centered
        # F face centered
        # A centered on A faces only
        # B centered on B faces only
        # C centered on C faces only
        # R rhombohedral
        # I grab this information from PyXtal's crystal.py file.
        if spacegroup in [
            22,
            42,
            43,
            69,
            70,
            196,
            202,
            203,
            209,
            210,
            216,
            219,
            225,
            226,
            227,
            228,
        ]:
            mult_prim = mult_conv / 4  # F
        elif spacegroup in [146, 148, 155, 160, 161, 166, 167]:
            mult_prim = mult_conv / 3  # R
        elif spacegroup in [
            5,
            8,
            9,
            12,
            15,
            20,
            21,
            23,
            24,
            35,
            36,
            37,
            38,
            39,
            40,
            41,
            44,
            45,
            46,
            63,
            64,
            65,
            66,
            67,
            68,
            71,
            72,
            73,
            74,
            79,
            80,
            82,
            87,
            88,
            97,
            98,
            107,
            108,
            109,
            110,
            119,
            120,
            121,
            122,
            139,
            140,
            141,
            142,
            197,
            199,
            204,
            206,
            211,
            214,
            217,
            220,
            229,
            230,
        ]:
            mult_prim = mult_conv / 2  # A, C, I
        else:
            mult_prim = mult_conv / 1  # P
        # append to output list
        wy_data.append([spacegroup, mult_conv, mult_prim, letter, sym, coords])

# close the window
driver.close()

#####

# make Asym DataFrame
df_asym = pd.DataFrame(
    data=asym_data, columns=["spacegroup", "asymmetricunit", "cellspec"]
)
# export to csv file
df_asym.to_csv("asymdata.csv", index=False)

#####

# make Asym DataFrame
df_wy = pd.DataFrame(
    data=wy_data,
    columns=[
        "SpaceGroup",
        "MultiplicityConventional",
        "MultiplicityPrimitive",
        "Wyckoffletter",
        "Sitesymmetry",
        "Coordinates",
    ],
)
# export to csv file
df_wy.to_csv("wyckoffdata.csv", index=False)

##############################################################################

# set this if you want to run headless (i.e. in the background)
# options = webdriver.ChromeOptions()
# options.headless = True
# and then add the setting to this call: webdriver.Chrome(options = options)


# customize the window if you'd like
# driver.minimize_window()
# driver.set_window_size(width=1920, height=1080)

##############################################################################



# -*- coding: utf-8 -*-

from simmate.workflow_engine import Workflow, task
from simmate.workflow_engine.common_tasks import load_input_and_register
from simmate.calculators.vasp.tasks.nudged_elastic_band import (
    MITNudgedElasticBand,
)
from simmate.calculators.vasp.database.nudged_elastic_band import (
    # MITDiffusionAnalysis,
    MITMigrationHop,
    MITMigrationImage,
)
from simmate.database.base_data_types.nudged_elastic_band import (
    # DiffusionAnalysis as DiffusionAnalysisTable,
    MigrationHop as MigrationHopTable,
    MigrationImage as MigrationImageTable,
)


class Diffusion__Vasp__NebFromImages(Workflow):
    """
    Runs a NEB relaxation on a list of structures (aka images) using MIT Project
    settings. The lattice remains fixed and symmetry is turned off for this
    relaxation.

    Typically, this workflow is not ran directly -- but instead, you would
    use diffusion/all-paths which submits a series of this workflow for you. Other
    higher level workflows that call this one include diffusion/from-endpoints
    and diffusion/single-path.
    """

    database_table = MITMigrationImage
    s3task = MITNudgedElasticBand
    description_doc_short = "runs NEB using a series of structures images as input"

    @classmethod
    def run_config(
        cls,
        migration_images,
        command: str = None,
        source: dict = None,
        directory: str = None,
        # These help link results to a higher-level table.
        diffusion_analysis_id: int = None,
        migration_hop_id: int = None,
        # TODO: Can the hop id be inferred from the migration_hop or somewhere
        # else in this context? Maybe even load_input_and_register will use
        # prefect id once it's a Calculation?
        is_restart: bool = False,
    ):

        # Load our input and make a base directory for all other workflows to run
        # within for us.
        parameters_cleaned = load_input_and_register(
            migration_images=migration_images,
            source=source,
            directory=directory,
            command=command,
            is_restart=is_restart,
            register_run=False,  # temporary fix bc no calc table exists yet
        )

        result = cls.s3task.run(**parameters_cleaned).result()

        calculation_id = save_neb_results(
            output=result,
            # migration_hop_table=MITDiffusionAnalysis,
            migration_hop_table=MITMigrationHop,
            migration_image_table=MITMigrationImage,
            diffusion_analysis_id=diffusion_analysis_id,
            migration_hop_id=migration_hop_id,
        )  # calculation_id will correspond to the migration hop table

        return result


@task
def save_neb_results(
    output,
    # diffusion_analyis_table: DiffusionAnalysisTable,
    migration_hop_table: MigrationHopTable,
    migration_image_table: MigrationImageTable,
    diffusion_analysis_id: int = None,
    migration_hop_id: int = None,
):

    # split our results and corrections (which are given as a dict) into
    # separate variables
    # Our result here is not a VaspRun object, but instead a NEBAnalysis
    # object. See NudgedElasticBandTask.workup()
    result = output["result"]

    # TODO: These aren't saved for now. Consider making MigrationHopTable
    # a Calculation and attaching these there.
    corrections = output["corrections"]
    directory = output["directory"]

    # First, we need a migration_hop database object.
    # All of hops should link to a diffusion_analysis entry, so we check
    # for that here too. The key thing of these statements is that we
    # have a migration_hop_id at the end.

    migration_hop_db = migration_hop_table.from_pymatgen(
        analysis=result,
        diffusion_analysis_id=diffusion_analysis_id,
        migration_hop_id=migration_hop_id,
    )

    # If the user wants to access results, they can do so through the hop id
    return migration_hop_db.id



# -*- coding: utf-8 -*-

"""
Workflows for running molecular dynamics simmulations
"""

# TODO: when I add more calculators, I can do something like this...
# if "simmate.calculators.vasp" in installed_apps:

from simmate.calculators.vasp.workflows.dynamics import Dynamics__Vasp__Mit



# -*- coding: utf-8 -*-

from simmate.website.core_components.filters import (
    Structure,
    Forces,
    Thermodynamics,
    Calculation,
)
from simmate.database.base_data_types import StaticEnergy as StaticEnergyTable


class StaticEnergy(
    Structure,
    Forces,
    Thermodynamics,
    Calculation,
):
    class Meta:
        model = StaticEnergyTable
        fields = dict(
            band_gap=["range"],
            is_gap_direct=["exact"],
            energy_fermi=["range"],
            conduction_band_minimum=["range"],
            valence_band_maximum=["range"],
            **Structure.get_fields(),
            **Forces.get_fields(),
            **Thermodynamics.get_fields(),
            **Calculation.get_fields(),
        )



# -*- coding: utf-8 -*-

"""
This module is experimental and subject to change.
"""


from simmate.database.base_data_types import (
    table_column,
    Structure,
    Forces,
    Thermodynamics,
    Calculation,
)

from typing import List
from pymatgen.io.vasp.outputs import Vasprun


class DynamicsRun(Structure, Calculation):
    """
    Holds results from a dynamics simulations -- often referred to as a molecular
    dynamics run.

    In addition to the attributes listed, you can also access all ionic steps
    of the run via the `structures` attribute. This attribute gives a list of
    `DynamicsIonicSteps`.
    """

    class Meta:
        abstract = True
        app_label = "workflows"

    temperature_start = table_column.IntegerField(blank=True, null=True)
    """
    The starting tempertature of the simulation in Kelvin.
    """

    temperature_end = table_column.IntegerField(blank=True, null=True)
    """
    The ending tempertature of the simulation in Kelvin.
    """

    time_step = table_column.FloatField(blank=True, null=True)
    """
    The time in picoseconds for each step in the simulation.
    """

    nsteps = table_column.IntegerField(blank=True, null=True)
    """
    The total number of steps in the simulation. Note, this is the maximum cutoff
    steps set. In some cases, there may be fewer steps when the simulation is 
    stopped early.
    """

    @classmethod
    def create_subclasses(cls, name: str, module: str, **extra_columns):
        """
        Dynamically creates a subclass of DynamicsRun as well as a separate
        DynamicsIonicStep table for it. These tables are linked together.

        Example use:

        ``` python
        from simmate.database.base_data_types import DynamicsRun

        ExampleDynamicsRun, ExampleDynamicsIonicStep = DynamicsRun.create_subclasses(
            "Example",
            module=__name__,
        )
        ```

        #### Parameters

        - `name` :
            The prefix name of the subclasses that are output. "DynamicsRun" and
            "DynamicsIonicStep" will be attached to the end of this prefix.
        - `module` :
            name of the module this subclass should be associated with. Typically,
            you should pass __name__ to this.
        - `**extra_columns` :
            Additional columns to add to the table. The keyword will be the
            column name and the value should match django options
            (e.g. table_column.FloatField())

        Returns
        -------
        - `NewDynamicsRunClass` :
            A subclass of DynamicsRun.
        - `NewDynamicsIonicStepClass`:
            A subclass of DynamicDynamicsIonicStep.

        """

        # For convience, we add columns that point to the start and end structures
        NewDynamicsRunClass = cls.create_subclass(
            f"{name}DynamicsRun",
            module=module,
            **extra_columns,
        )

        NewDynamicDynamicsIonicStepClass = (
            DynamicsIonicStep.create_subclass_from_dynamics_run(
                name,
                NewDynamicsRunClass,
                module=module,
                **extra_columns,
            )
        )

        # we now have a new child class and avoided writing some boilerplate code!
        return NewDynamicsRunClass, NewDynamicDynamicsIonicStepClass

    def update_from_vasp_run(
        self,
        vasprun: Vasprun,
        corrections: List,
        directory: str,
    ):
        """
        Given a Vasprun object from a finished dynamics run, this will update the
        DynamicsRun table entry and the corresponding DynamicsIonicStep entries.

        #### Parameters

        vasprun :
            The final Vasprun object from the dynamics run outputs.
        corrections :
            List of errors and corrections applied to during the relaxation.
        directory :
            name of the directory that relaxation was ran in. This is only used
            to reference the archive file if it's ever needed again.
        """

        # The data is actually easier to access as a dictionary and everything
        # we need is stored under the "output" key
        data = vasprun.as_dict()["output"]

        # The only other data we need to grab is the list of structures. We can
        # pull the structure for each ionic step from the vasprun class directly.
        structures = vasprun.structures

        # Now let's iterate through the ionic steps and save these to the database.
        for number, (structure, ionic_step) in enumerate(
            zip(structures, data["ionic_steps"])
        ):
            # first pull all the data together and save it to the database. We
            # are saving this to an DynamicsIonicStepStructure datatable. To access this
            # model, we look need to use "structures.model".
            structure = self.structures.model.from_toolkit(
                number=number,
                structure=structure,
                energy=ionic_step.get("e_wo_entrp", None),
                site_forces=ionic_step.get("forces", None),
                lattice_stress=ionic_step.get("stress", None),
                temperature=self._get_temperature_at_step(number),
                dynamics_run=self,  # this links the structure to this dynamics run
            )
            structure.save()

        # lastly, we also want to save the corrections made and directory it ran in
        self.corrections = corrections
        self.directory = directory

        # Now we have the relaxation data all loaded and can save it to the database
        self.save()

    def _get_temperature_at_step(self, step_number: int):
        return step_number * self._get_temperature_step_size() + self.temperature_start

    def _get_temperature_step_size(self):
        return (self.temperature_end - self.temperature_start) / self.nsteps


class DynamicsIonicStep(Structure, Thermodynamics, Forces):
    """
    Holds information for a single ionic step of a `DynamicsRun`.

    Each entry will map to a `DynamicsRun`, so you should typically access this
    data through that class. The exception to this is when you want all ionic
    steps accross many relaxations for a machine learning input.
    """

    class Meta:
        abstract = True
        app_label = "workflows"

    base_info = (
        ["number"] + Structure.base_info + Thermodynamics.base_info + Forces.base_info
    )

    number = table_column.IntegerField()
    """
    This is ionic step number for the given relaxation. This starts counting from 0.
    """

    temperature = table_column.FloatField(blank=True, null=True)
    """
    Expected temperature based on the temperature_start/end and nsteps. Note
    that in-practice some steps may not be at equilibrium temperatue. This could
    be the 1st 100 steps of a run or alternatively when the thermostat component
    is off-temperature.
    """

    # TODO: Additional options from Vasprun.as_dict to consider adding
    # e_0_energy
    # e_fr_energy
    # kinetic
    # lattice kinetic
    # nosekinetic
    # nosepot

    @classmethod
    def create_subclass_from_dynamics_run(
        cls,
        name: str,
        dynamics_run: DynamicsRun,
        module: str,
        **extra_columns,
    ):
        """
        Dynamically creates a subclass of DynamicsIonicStep and links it to the
        DynamicsRun table.

        This method should NOT be called directly because it is instead used by
        `DynamicsRun.create_subclasses`.

        #### Parameters

        - `name` :
            Name of the subclass that is output.
        - `dynamics_run` :
            DynamicsRun table that these ionic steps should be associated with.
        - `module` :
            name of the module this subclass should be associated with. Typically,
            you should pass __name__ to this.
        - `**extra_columns` :
            Additional columns to add to the table. The keyword will be the
            column name and the value should match django options
            (e.g. table_column.FloatField())

        Returns
        -------
        - `NewClass` :
            A subclass of DynamicsIonicStep.

        """

        # All structures in this table come from dynamics run calculations, where
        # there can be many structures (one for each ionic step) linked to a
        # single run. This means the start structure, end structure, and
        # those structure in-between are stored together here.
        # Therefore, there's just a simple column stating which relaxation it
        # belongs to.
        NewClass = cls.create_subclass(
            f"{name}DynamicsIonicStep",
            dynamics_run=table_column.ForeignKey(
                dynamics_run,
                on_delete=table_column.CASCADE,
                related_name="structures",
            ),
            module=module,
            **extra_columns,
        )

        # we now have a new child class and avoided writing some boilerplate code!
        return NewClass



# -*- coding: utf-8 -*-

from simmate.workflow_engine import Workflow
from simmate.calculators.vasp.tasks.static_energy import (
    MITStaticEnergy as MITStaticEnergyTask,
)
from simmate.calculators.vasp.database.energy import (
    MITStaticEnergy as MITStaticEnergyResults,
)


class StaticEnergy__Vasp__Mit(Workflow):
    s3task = MITStaticEnergyTask
    database_table = MITStaticEnergyResults
    description_doc_short = "uses MIT Project settings"



# -*- coding: utf-8 -*-

import os

from simmate.conftest import copy_test_files
from simmate.calculators.vasp.tasks.density_of_states import MatprojDensityOfStates


def test_density_of_states_workup(tmpdir):
    copy_test_files(
        tmpdir,
        test_directory=__file__,
        test_folder="density_of_states",
    )

    # init main task
    task = MatprojDensityOfStates()

    # estabilish filenames that we make and commonly reference
    summary_filename = os.path.join(tmpdir, "simmate_summary.yaml")
    plot_filename = os.path.join(tmpdir, "density_of_states.png")

    # run the full workup
    task.workup(tmpdir)
    assert os.path.exists(summary_filename)
    assert os.path.exists(plot_filename)



# -*- coding: utf-8 -*-


class StdErrHandler(ErrorHandler):
    """
    Master StdErr class that handles a number of common errors
    that occur during VASP runs with error messages only in
    the standard error.
    """

    is_monitor = True

    error_msgs = {
        "kpoints_trans": [
            "internal error in GENERATE_KPOINTS_TRANS: "
            "number of G-vector changed in star"
        ],
        "out_of_memory": ["Allocation would exceed memory limit"],
    }

    def __init__(self, output_filename="std_err.txt"):
        """
        Initializes the handler with the output file to check.
        Args:
            output_filename (str): This is the file where the stderr for vasp
                is being redirected. The error messages that are checked are
                present in the stderr. Defaults to "std_err.txt", which is the
                default redirect used by :class:`custodian.vasp.jobs.VaspJob`.
        """
        self.output_filename = output_filename
        self.errors = set()
        self.error_count = Counter()

    def check(self):
        """
        Check for error.
        """
        self.errors = set()
        with open(self.output_filename, "r") as f:
            for line in f:
                l = line.strip()
                for err, msgs in StdErrHandler.error_msgs.items():
                    for msg in msgs:
                        if l.find(msg) != -1:
                            self.errors.add(err)
        return len(self.errors) > 0

    def correct(self):
        """
        Perform corrections.
        """
        backup(VASP_BACKUP_FILES | {self.output_filename})
        actions = []
        vi = VaspInput.from_directory(".")

        if "kpoints_trans" in self.errors:
            if self.error_count["kpoints_trans"] == 0:
                m = reduce(operator.mul, vi["KPOINTS"].kpts[0])
                m = max(int(round(m ** (1 / 3))), 1)
                if vi["KPOINTS"].style.name.lower().startswith("m"):
                    m += m % 2
                actions.append(
                    {"dict": "KPOINTS", "action": {"_set": {"kpoints": [[m] * 3]}}}
                )
                self.error_count["kpoints_trans"] += 1

        if "out_of_memory" in self.errors:
            if vi["INCAR"].get("KPAR", 1) > 1:
                reduced_kpar = max(vi["INCAR"].get("KPAR", 1) // 2, 1)
                actions.append(
                    {"dict": "INCAR", "action": {"_set": {"KPAR": reduced_kpar}}}
                )

        VaspModder(vi=vi).apply_actions(actions)
        return {"errors": list(self.errors), "actions": actions}



# -*- coding: utf-8 -*-

import os

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class IncorrectShift(ErrorHandler):
    """
    This handler addresses issues in the K-point mesh that can be fixed by switching
    to a gamma-centered mesh.
    """

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = ["Could not get correct shifts"]

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # Change to Gamma-centered mesh
        incar["KGAMMA"] = True

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return "switched KGAMMA to True"



# -*- coding: utf-8 -*-

"""
This is where you add any new database tables. This file is named
"models.py" because that what Django expects! In their terminology, a django 
model is the same thing as a database table.

Whenever building new tables, be sure to start from the classes located in our
`simmate.database.base_data_types` module. These let you automatically add useful 
columns and features -- rather than creating everything from scratch. 

For more information and advanced guides, be sure to read through our
[base_data_types documentation](https://jacksund.github.io/simmate/simmate/database/base_data_types.html)
"""

from simmate.database.base_data_types import DatabaseTable, Relaxation, table_column


# As an example here, we set up tables for a relaxation. There are two tables
# here: one for the relaxations and a second for storing each ionic step in the
# relaxation. Both of these tables are created automatically with the create_subclasses
# method:
ExampleRelaxation, ExampleIonicStep = Relaxation.create_subclasses(
    "Example",
    module=__name__,  # this line is required
)

# alternatively, we can make a DatabaseTable from scratch
class MyCustomTable(DatabaseTable):
    custom_column_01 = table_column.IntegerField()
    custom_column_02 = table_column.FloatField(null=True, blank=True)



# -*- coding: utf-8 -*-

import os

from pymatgen.io.vasp.outputs import Vasprun
from pymatgen.electronic_structure.dos import CompleteDos
from pymatgen.electronic_structure.plotter import DosPlotter

from simmate.database.base_data_types import (
    table_column,
    DatabaseTable,
    Calculation,
    Structure,
)


class DensityofStates(DatabaseTable):
    """
    The electronic density of states holds information about the range of energy
    levels that are available within a material.
    """

    class Meta:
        abstract = True

    base_info = ["density_of_states_data"]

    # !!! Consider breaking down data into...
    #   total
    #   elemental -- dict of ["total", "s", "p", "d", "f"], ["1", "-1"]
    #   orbital
    density_of_states_data = table_column.JSONField(blank=True, null=True)
    """
    A JSON dictionary holding all information for the band structure. This JSON
    is generated using pymatgen's `vasprun.complete_dos.as_dict()` and is 
    therefore currently unoptimized for small storage.
    """

    band_gap = table_column.FloatField(blank=True, null=True)
    """
    The band gap energy in eV.
    """

    energy_fermi = table_column.FloatField(blank=True, null=True)
    """
    The Fermi energy in eV.
    """

    conduction_band_minimum = table_column.FloatField(blank=True, null=True)
    """
    The conduction band minimum in eV.
    """

    valence_band_maximum = table_column.FloatField(blank=True, null=True)
    """
    The valence band maximum in eV.
    """

    # TODO: consider adding...
    # spin_polarization (float "Spin polarization at the fermi level.")
    # magnetic_ordering (Magnetic ordering of the calculation.)

    @classmethod
    def _from_toolkit(
        cls,
        density_of_states: CompleteDos = None,
        as_dict=False,
    ):
        # Given energy, this function builds the rest of the required fields
        # for this class as an object (or as a dictionary).
        data = (
            dict(
                density_of_states_data=density_of_states.as_dict(),
                band_gap=density_of_states.get_gap(),
                energy_fermi=density_of_states.efermi,
                conduction_band_minimum=density_of_states.get_cbm_vbm()[0],
                valence_band_maximum=density_of_states.get_cbm_vbm()[1],
            )
            if density_of_states
            else {}
        )

        # If as_dict is false, we build this into an Object. Otherwise, just
        # return the dictionary
        return data if as_dict else cls(**data)

    def to_toolkit_density_of_states(self) -> CompleteDos:
        """
        Converts this DatabaseTable object into a toolkit CompleteDos, which
        has many more methods for plotting and analysis.
        """
        return CompleteDos.from_dict(self.density_of_states_data)

    def get_densityofstates_plot(self):
        """
        Plots the density of states using matplotlib
        """

        # NOTE: This method should be moved to a toolkit object

        # DEV NOTE: Pymatgen only implements matplotlib for their DOS
        # at the moment, but there are two scripts location elsewhere that can
        # outline how this can be done with Plotly:
        # https://plotly.com/python/v3/ipython-notebooks/density-of-states/
        # https://github.com/materialsproject/crystaltoolkit/blob/main/crystal_toolkit/components/bandstructure.py

        plotter = DosPlotter()
        complete_dos = self.to_toolkit_density_of_states()

        # Add the total density of States
        plotter.add_dos("Total DOS", complete_dos)

        # add element-projected density of states
        plotter.add_dos_dict(complete_dos.get_element_dos())

        # If I want plots for individual orbitals
        # for site in vasprun.final_structure:
        #     spd_dos = vasprun.complete_dos.get_site_spd_dos(site)
        #     plotter.add_dos_dict(spd_dos)

        # NOTE: get_dos_dict may be useful in the future

        plot = plotter.get_plot()
        return plot


class DensityofStatesCalc(Structure, DensityofStates, Calculation):
    """
    Holds Structure, DensityofStates, and Calculation information. Density of state
    workflows are common in materials science, so this table defines the most
    common data that results from such workflow calculations.
    """

    class Meta:
        abstract = True
        app_label = "workflows"

    base_info = Structure.base_info + DensityofStates.base_info + Calculation.base_info

    def update_from_vasp_run(self, vasprun: Vasprun, corrections: list, directory: str):
        """
        Given a pymatgen VaspRun object, which is what's typically returned
        from a simmate VaspTask.run() call, this will update the database entry
        with the results.
        """

        # Takes a pymatgen VaspRun object, which is what's typically returned
        # from a simmate VaspTask.run() call.

        # All data analysis is done via a CompleteDOS object, so we convert
        # the vasprun object to that first.
        density_of_states = vasprun.complete_dos

        # Take our dos and expand its data for the rest of the columns.
        new_kwargs = DensityofStates.from_toolkit(
            density_of_states=density_of_states,
            as_dict=True,
        )
        for key, value in new_kwargs.items():
            setattr(self, key, value)

        # lastly, we also want to save the corrections made and directory it ran in
        self.corrections = corrections
        self.directory = directory

        # Now we have the relaxation data all loaded and can save it to the database
        self.save()

    @classmethod
    def from_directory(cls, directory: str):
        """
        Creates a new database entry from a directory that holds band-structure
        results. For now, this assumes the directory holds vasp output files.
        """

        # I assume the directory is from a vasp calculation, but I need to update
        # this when I begin adding new calculators.
        vasprun_filename = os.path.join(directory, "vasprun.xml")
        vasprun = Vasprun(vasprun_filename)
        density_of_states_db = cls.from_toolkit(
            structure=vasprun.structures[0],
            density_of_states=vasprun.complete_dos,
        )
        density_of_states_db.save()
        return density_of_states_db



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.relaxation import MatprojRelaxation


class MatprojStaticEnergy(MatprojRelaxation):
    """
    This task is a reimplementation of pymatgen's
    [MPStaticSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPStaticSet).

    Runs a VASP static energy calculation using Materials Project settings.

    This is identical to relaxation/Matproj, but just a single ionic step.
    """

    incar = MatprojRelaxation.incar.copy()
    incar.update(
        dict(
            IBRION=-1,  # (optional) locks everything between ionic steps
            NSW=0,  # this is the main static energy setting
            LAECHG=True,
            LCHARG=True,
            LORBIT=11,
            LVHAR=True,
            LWAVE=False,
            ALGO="Normal",  # was "Fast" before
        )
    )



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.nudged_elastic_band.mit import MITNudgedElasticBand


class MatVirtualLabClimbingImageNudgedElasticBand(MITNudgedElasticBand):
    """
    This task is a reimplementation of pymatgen's
    [MVLCINEBSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MITNEBSet).

    Runs a NEB relaxation on a list of structures (aka images) using MIT Project
    settings. The lattice remains fixed and symmetry is turned off for this
    relaxation.

    You shouldn't use this workflow directly, but instead use the higher-level
    NEB workflows (e.g. diffusion/neb_all_paths or diffusion/neb_from_endpoints),
    which call this workflow for you.

    Note that these parameters require the VTST modification of VASP from the
    Henkelman group. See http://theory.cm.utexas.edu/vtsttools/
    """

    incar = MITNudgedElasticBand.incar.copy()
    incar.update(
        dict(
            EDIFF=5e-5,
            EDIFFG=-0.02,
            IBRION=3,
            ICHAIN=0,
            IOPT=1,
            ISIF=2,
            ISMEAR=0,
            LCLIMB=True,
            LORBIT=0,
            NSW=200,
            POTIM=0,
            SPRING=-5,
        )
    )



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import table_column, Structure


class JarvisStructure(Structure):
    """
    Crystal structures from the [JARVIS](https://jarvis.nist.gov/) database.

    Currently, this table only stores strucure and reported energy above hull.
    The calculated energy is not reported, so the Thermodynamics mixin is not used.
    """

    class Meta:
        app_label = "third_parties"

    base_info = ["id", "structure_string", "energy_above_hull"]
    source = "JARVIS"
    source_doi = "https://doi.org/10.1038/s41524-020-00440-1"
    remote_archive_link = "https://archives.simmate.org/JarvisStructure-2022-01-26.zip"

    id = table_column.CharField(max_length=25, primary_key=True)
    """
    The id used to represent the structure (ex: "jvasp-12345")
    """

    # TODO: contact their team to ask about reporting energy instead. That way
    # we can use the Thermodynamics mixin instead of manually listing this.
    energy_above_hull = table_column.FloatField(blank=True, null=True)
    """
    The energy above hull, as reported by the JARVIS database (no units given)
    """

    @property
    def external_link(self) -> str:
        """
        URL to this structure in the JARVIS website.
        """
        # All JARVIS structures have their data mapped to a URL in the same way
        # ex: https://www.ctcms.nist.gov/~knc6/static/JARVIS-DFT/JVASP-1234.xml
        # we store the id as "jvasp-123" so we need to convert this to uppercase
        id = self.id.upper()
        return f"https://www.ctcms.nist.gov/~knc6/static/JARVIS-DFT/{id}.xml"



# -*- coding: utf-8 -*-

from django_filters import rest_framework as filters

from simmate.website.core_components.filters import Calculation, Structure
from simmate.database.base_data_types import (
    DensityofStates as DensityofStatesTable,
    DensityofStatesCalc as DensityofStatesCalcTable,
)


class DensityofStates(filters.FilterSet):
    class Meta:
        model = DensityofStatesTable
        fields = dict(
            band_gap=["range"],
            energy_fermi=["range"],
            conduction_band_minimum=["range"],
            valence_band_maximum=["range"],
        )


class DensityofStatesCalc(DensityofStates, Calculation, Structure):
    class Meta:
        model = DensityofStatesCalcTable
        fields = {
            **Structure.get_fields(),
            **DensityofStates.get_fields(),
            **Calculation.get_fields(),
        }



# -*- coding: utf-8 -*-

import pytest

from pandas import DataFrame

from simmate.toolkit import Structure
from simmate.website.test_app.models import TestStructure


@pytest.mark.django_db
def test_structure_table(structure):

    # test writing columns
    TestStructure.show_columns()

    # test writing to database
    structure_db = TestStructure.from_toolkit(structure=structure)
    structure_db.save()

    # test writing to dictionary
    structure_dict = TestStructure.from_toolkit(structure=structure, as_dict=True)
    assert isinstance(structure_dict, dict)

    # test converting back to toolkit and ensuring the structure is the
    # same as before
    structure_new = structure_db.to_toolkit()
    assert structure == structure_new


@pytest.mark.django_db
def test_structure_queries():

    # test converting search results to dataframe and to toolkit
    df = TestStructure.objects.to_dataframe()
    assert isinstance(df, DataFrame)

    structures = TestStructure.objects.to_toolkit()
    assert isinstance(structures, list)
    assert isinstance(structures[0], Structure)


@pytest.mark.django_db
def test_structure_archives():

    TestStructure.objects.to_archive()
    TestStructure.load_archive(
        confirm_override=True,
        delete_on_completion=True,
    )



# -*- coding: utf-8 -*-

import math

from simmate.toolkit import Structure

# TODO: consider making modifiers a class. For example:
#
# class PerAtom(KeywordModifier):
#     tag_name = "__per_atom"
#     @staticmethod
#     def apply(structure, per_atom_value):
#         return per_atom_value * structure.num_sites


def keyword_modifier_density(structure: Structure, density: float):
    """
    The __density modifier means the user wants a specific density. They
    provide this density in per-angstrom^3 units and we return the
    structure-specific count that gives this density.
    For example, density=10 and a structure lattice that volume of 5,
    then this returns value=10*5=50.
    """
    # VASP expect integers for a lot of these values, so we round up
    return math.ceil(structure.lattice.volume * density)


def keyword_modifier_density_a(structure: Structure, density: float):
    """
    The __density_a modifier means the user wants a specific density along
    the A lattice vector. They provide this density in per-angstrom units
    and we return the structure-specific count that gives this density.
    For example, density=10 and a structure lattice that A vector of 5,
    then this returns value=10*5=50.
    """
    # VASP expect integers for a lot of these values, so we round up
    return math.ceil(structure.lattice.a * density)


def keyword_modifier_density_b(structure: Structure, density: float):
    return math.ceil(structure.lattice.b * density)


def keyword_modifier_density_c(structure: Structure, density: float):
    return math.ceil(structure.lattice.c * density)


def keyword_modifier_per_atom(structure: Structure, per_atom_value: float):
    """
    The __per_atom modifier means the user wants a specific value per atom
    in the unit cell. For example, EDIFF__per_atom=1e-5 and a structure
    with 50 sites in it would return a value of 1e-5*50=50.
    """
    return per_atom_value * structure.num_sites


def keyword_modifier_smart_magmom(structure: Structure, override_values: dict):
    """
    The __smart_magmom modifier goes through a series of checks to decide
    what to set the MAGMOM as for VASP. In order of priority they are...
        (1) the magmom property attached to each site in the structure object
        (2) the spin property attached to the site's specie
        (3) a value provided explicitly (e.g. {"Co": 0.5})
        (4) a value of 0.6
    """

    # grab the default MAGMOM supplied, or use VASP's default of 1 otherwise
    default_value = override_values.get("default", 1)

    # we go through each site in the structure and decide what to set the
    # MAGMOM for each. This allows even different sites of the same
    # element to have their own MAGMOM
    magnetic_moments = []
    for site in structure:
        # if the structure object has magmom-decorated sites, we use that
        # as our first priority
        if hasattr(site, "magmom"):
            magnetic_moments.append(site.magmom)
        # next we check if the site's specie has a spin
        elif hasattr(site.specie, "spin"):
            magnetic_moments.append(site.specie.spin)
        # we then look at the override dictionary if there was one provided.
        # If note, we use 0.6 as a default.
        else:
            magnetic_moment = override_values.get(site.specie.symbol, default_value)
            magnetic_moments.append(magnetic_moment)

    # This feature is in pymatgen, but I haven't added it here yet.
    # if self.constrain_total_magmom:
    #     nupdown = sum([mag if abs(mag) > 0.6 else 0 for mag in incar["MAGMOM"]])
    #     incar["NUPDOWN"] = nupdown

    return magnetic_moments


def keyword_modifier_smart_lmaxmix(structure: Structure, lmaxmix_config: bool):
    """
    This modifier sets LMAXMIX if there are any d or f electrons present
    in the structure.
    """

    # COMMENT (from pymatgen team):
    # Note that if the user explicitly sets LMAXMIX in settings it will
    # override this logic.
    # Previously, this was only set if Hubbard U was enabled as per the
    # VASP manual but following an investigation it was determined that
    # this would lead to a significant difference between SCF -> NonSCF
    # even without Hubbard U enabled. Thanks to Andrew Rosen for
    # investigating and reporting.

    # first iterate through all elements and check for f-electrons
    if any(element.Z > 56 for element in structure.composition):
        return 6
    # now check for elements that contain d-electrons
    elif any(element.Z > 20 for element in structure.composition):
        return 4
    # otherwise use the default for VASP
    else:
        return None  # default is 2, but we don't set it


def keyword_modifier_smart_ldau(structure: Structure, ldau_config: dict):
    """
    This modifier handles a series of keyword arguments that are associated
    with LDAU, including LDAUJ, LDAUL, LDAUTYPE, LDAUU, and LDAUPRINT.
    Therefore, a complex dictionary is passed to this. The format looks
    like this...
        LDAU__multiple_keywords__smart_ldau = dict(
            LDAU__auto=True,
            LDAUTYPE=2,
            LDAUPRINT=1,
            LDAUJ={...},
            LDAUL={...},
            LDAUU={...},
            )
    The LDAUJ, LDAUL, and LDAUU values can be a dictionary of elements to
    value, or (most commonly) a nested dictionary. For example...
        LDAUJ = {"F":{"Co":0}}
    This would mean if the structure is a fluoride, set the LDAUJ for Co to 0.
    If there are multiple options here (e.g. for fluorides and oxides), then
    priority is placed on the most electronegative element. This would mean
    some thing like yttrium oxide fluoride would be treated as a fluoride over
    an oxide.
    """

    # first we need to go through the LDAUJ, LDAUL, and LDAUU keywords and
    # see what their values are. If all of these end up be 0 for all elements
    # then we actually don't need LDAU at all! Therefore, we'll go through
    # all of these keywords and build a dictionary of settings to return
    # at the end of this function
    ldau_settings = {}

    # To help decide how we set these values, let's check what the most
    # electronegative element is, which will be last in the sorted composition
    most_electroneg = sorted(structure.composition.elements)[-1].symbol

    # As we go through these settings, we want see if we are even using LDAU.
    # For example if we ran a calculation on NaCl, we'd probably see that
    # all LDAUJ/LDAUL/LDAUU values are just 0. In that case, we can just
    # throw away (i.e. turn off) all LDAU settings. Therefore, we have a term
    # "using_ldau" that is false until proven otherwise.
    using_ldau = False

    for ldau_keyword in ["LDAUJ", "LDAUL", "LDAUU"]:

        # grab the sub-dictionary that maps elements to this keyword.
        # If it's not there, just use an empty dictionary.
        keyword_config = ldau_config.get(ldau_keyword, {})

        # check if the most electronegative element is in the override_values
        # and if so, see if it has a subdictionary in it. We use this
        # dictionary as our base one to pull values from.
        if most_electroneg in keyword_config and isinstance(
            keyword_config[most_electroneg], dict
        ):
            keyword_config = keyword_config[most_electroneg]

        # now iterate through all the sites and grab the assigned value. If nothing
        # is set then the default is 0.
        values = [
            keyword_config.get(element.symbol, 0) for element in structure.composition
        ]

        # now that we have this keyword all set, we add it to our results
        ldau_settings[ldau_keyword] = values

        # check to see if we are using ldau here (i.e. any value is not 0)
        if any(values):
            using_ldau = True

    # now check if we actaully need LDAU here. If not, we can throw out all
    # settings and just return an empty dictionary
    if using_ldau and "LDAU__auto" in ldau_config:
        ldau_settings["LDAU"] = True
    else:
        return {}

    # The remaining LDAU keywords are LDAUPRINT and LDAUTYPE, which we just
    # leave at what is set in the input
    if "LDAUPRINT" in ldau_config:
        ldau_settings["LDAUPRINT"] = ldau_config["LDAUPRINT"]
    if "LDAUTYPE" in ldau_config:
        ldau_settings["LDAUTYPE"] = ldau_config["LDAUTYPE"]

    return ldau_settings


def keyword_modifier_smart_ismear(structure: Structure, ismear_config: dict):
    """
    The smearing value used here depends on if we have a semiconductor,
    insulator, or metal. This modifier makes a "best-guess" on what the
    material is and uses the proper smearing type. Note that if this
    guess is wrong, it is useful to have the IncorrectSmearing error
    handler to fix this as VASP runs.

    Read more about the VASP recommended ISMEAR settings here:
        https://www.vasp.at/wiki/index.php/ISMEAR
    """

    # for now we just go through the structure and if all elements are
    # metals, then we say it's a metal. Otherwise, we treat the structure
    # as a semiconductor or insulator.
    if all(element.is_metal for element in structure.composition):
        ismear_settings = ismear_config.get("metal", {})

    else:
        ismear_settings = ismear_config.get("non-metal", {})

    return ismear_settings


# TODO: In the future, I want to allow modifiers like __relative_to_previous
# and __use_previous to string settings accross tasks.

# To introduce other modifiers that pymatgen uses...
# https://github.com/materialsproject/pymatgen/blob/b789d74639aa851d7e5ee427a765d9fd5a8d1079/pymatgen/io/vasp/sets.py#L500

# if self.use_structure_charge:
#     incar["NELECT"] = self.nelect

# Ensure adequate number of KPOINTS are present for the tetrahedron
# method (ISMEAR=-5). If KSPACING is in the INCAR file the number
# of kpoints is not known before calling VASP, but a warning is raised
# when the KSPACING value is > 0.5 (2 reciprocal Angstrom).
# An error handler in Custodian is available to
# correct overly large KSPACING values (small number of kpoints)
# if necessary.
# if "KSPACING" not in self.user_incar_settings.keys():
# if self.kpoints is not None:
#     if np.product(self.kpoints.kpts) < 4 and incar.get("ISMEAR", 0) == -5:
#         incar["ISMEAR"] = 0



# -*- coding: utf-8 -*-

import os

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class Aliasing(ErrorHandler):
    """
    Aliasing errors occur when there are issues with the NGX, NGY, and NGZ grid
    in VASP. Typically the fix is to just remove user defined settings for these
    values, delete the CHGCAR and WAVECAR, and then restart the calculation. In
    some cases, VASP even provides recommended values for this grid and we just
    use those.
    """

    # run this while the VASP calculation is still going
    is_monitor = True

    # These are the error messages that we are looking for
    possible_error_messages = {
        "aliasing_error": "WARNING: small aliasing (wrap around) errors must be expected",
        "insufficient_fft_grid": "Your FFT grids (NGX,NGY,NGZ) are not sufficient for an accurate",
    }

    # All instances of this ErrorHandler will work exactly the same, so there
    # are no keywords available -- and also there's no need for a __init__()

    def check(self, directory):
        """
        Check for errors in the specified directory. Note, we assume that we are
        checking the vasp.out file. If that file is not present, we say that there
        is no error because another handler will address this.
        """

        # We want to return a list of errors that were found, so we keep a
        # master list.
        errors_found = []

        # establish the full path to the output file
        filename = os.path.join(directory, "vasp.out")

        # check to see that the file is there first
        if os.path.exists(filename):

            # read the file content and then close it
            with open(filename) as file:
                file_text = file.read()

            # Check if each error is present
            for error, message in self.possible_error_messages.items():
                # if the error is NOT present, find() returns a -1
                if file_text.find(message) != -1:
                    errors_found.append(error)

        # If the file doesn't exist, we are not seeing any error yetm which is
        # also an empty list. Otherwise return the list of errors we found
        return errors_found

    def correct(self, directory):
        """
        Perform corrections based on the INCAR.
        """

        # Note "error" here is a list of the errors found. For example, error
        # could be ["alaising_error", "insufficient_fft_grid"]. If there
        # weren't any errors found then this is just and empty list.

        # Multiple corrections could be applied here so we record all of them.
        corrections = []

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        if "aliasing" in error:

            # VASP will actually tell us what to switch NGX, NGY, and NGZ to inside
            # the OUTCAR. So we look there for our fix.
            # Here is what the aliasing error typically looks like in the OUTCAR:
            #   WARNING: aliasing errors must be expected set NGX to  34 to avoid them
            #   NGY is ok and might be reduce to 102
            #   NGZ is ok and might be reduce to 120
            #   aliasing errors are usually negligible using standard VASP settings
            #   and one can safely disregard these warnings

            # so we open the OUTCAR and find these lines and the suggested fix.

            # read the file content and then close it
            outcar_filename = os.path.join(directory, "OUTCAR")
            with open(outcar_filename) as file:
                outcar_lines = file.readlines()

            # go through each line looking for the suggestions
            for line in outcar_lines:
                if "to avoid them" in line:

                    # break the line up into a list, where NG* is at index 7 and
                    # the fix is at index 9.
                    # BUG: I expect an error will show up here if I was wrong about
                    # the indexes of what we want always being 7 and 9.
                    line_info = line.split()
                    incar_key = line_info[7]
                    new_value = int(line_info[9])

                    # update the incar with this value
                    incar.update({incar_key: new_value})

                    # record the correction
                    corrections.append(f"switched {incar_key} to {new_value}")

        if "insufficient_fft_grid" in error:

            # This is typically caused by a user introducing too coarse of a grid.
            # Here we remove all NGX/NGY/NGZ settings from the INCAR
            incar.pop("NGX")
            incar.pop("NGY")
            incar.pop("NGZ")
            # record the change
            corrections.append("removed all NGX/NGY/NGZ settings")

        # if any corrections were made above, we may want to reset some files
        if corrections:

            # Check the current ICHARG setting, where default is 0
            # !!! BUG: isn't the default 2 if we are starting from scratch?
            current_icharg = incar.get("ICHARG", 0)

            # If the ICHARG is less than 10, then we want to delete the CHGCAR
            # and WAVECAR to ensure the next run is a clean start.
            if current_icharg < 10:
                os.remove(os.path.join(directory, "CHGCAR"))
                os.remove(os.path.join(directory, "WAVECAR"))

            # return the description of what we did
            corrections.append("deleted CHGCAR and WAVECAR")

        # rewrite the new INCAR file
        incar.to_file(incar_filename)

        # return the list of corrections we made
        return corrections



# -*- coding: utf-8 -*-

from .matproj import Relaxation__Vasp__Matproj
from .mit import Relaxation__Vasp__Mit
from .neb_endpoint import Relaxation__Vasp__NebEndpoint
from .quality00 import Relaxation__Vasp__Quality00
from .quality01 import Relaxation__Vasp__Quality01
from .quality02 import Relaxation__Vasp__Quality02
from .quality03 import Relaxation__Vasp__Quality03
from .quality04 import Relaxation__Vasp__Quality04
from .staged import Relaxation__Vasp__Staged



# -*- coding: utf-8 -*-


class ScanMetalHandler(ErrorHandler):
    """
    Check if a SCAN calculation is a metal (zero bandgap) but has been run with
    a KSPACING value appropriate for semiconductors. If this occurs, this handler
    will rerun the calculation using the KSPACING setting appropriate for metals
    (KSPACING=0.22). Note that this handler depends on values set in MPScanRelaxSet.
    """

    is_monitor = False

    def __init__(self, output_filename="vasprun.xml"):
        """
        Initializes the handler with the output file to check.
        Args:
            output_filename (str): Filename for the vasprun.xml file. Change
                this only if it is different from the default (unlikely).
        """
        self.output_filename = output_filename

    def check(self):
        """
        Check for error.
        """
        try:
            v = Vasprun(self.output_filename)
            # check whether bandgap is zero and tetrahedron smearing was used
            if (
                v.eigenvalue_band_properties[0] == 0
                and v.incar.get("KSPACING", 1) > 0.22
            ):
                return True
        except Exception:
            pass
        return False

    def correct(self):
        """
        Perform corrections.
        """
        backup(VASP_BACKUP_FILES | {self.output_filename})
        vi = VaspInput.from_directory(".")

        _dummy_structure = Structure(
            [1, 0, 0, 0, 1, 0, 0, 0, 1],
            ["I"],
            [[0, 0, 0]],
        )
        new_vis = MPScanRelaxSet(_dummy_structure, bandgap=0)

        actions = []
        actions.append(
            {
                "dict": "INCAR",
                "action": {"_set": {"KSPACING": new_vis.incar["KSPACING"]}},
            }
        )

        VaspModder(vi=vi).apply_actions(actions)
        return {"errors": ["ScanMetal"], "actions": actions}



# -*- coding: utf-8 -*-

import numpy
from tqdm import tqdm

from django.utils import timezone

# BUG: This prints a tqdm error so we silence it here.
import warnings

with warnings.catch_warnings(record=True):
    from matminer.featurizers.site import CrystalNNFingerprint

# BUG: waiting on this pSS to be added to matminer. See the following PR:
#   https://github.com/hackingmaterials/matminer/pull/809
from matminer.featurizers.structure.sites import (
    SiteStatsFingerprint,
)  # PartialsSiteStatsFingerprint


class PartialCrystalNNFingerprint:
    def __init__(
        self,
        composition,
        stat_options=["mean", "std_dev", "minimum", "maximum"],
        structure_pool=[],  # either a list of structures OR a queryset
        add_unique_to_pool=True,  # whether to add fingerprint when check_structure is true
        **crystalnn_options,
    ):

        self.add_unique_to_pool = add_unique_to_pool

        # make the matminer featurizer object using the provided settings
        if crystalnn_options:
            sitefingerprint_method = CrystalNNFingerprint(**crystalnn_options)
        # if no settings were given, we will use those from the Materials Project
        #   https://docs.materialsproject.org/user-guide/structure-similarity/
        else:
            # We change x_diff_weight to 3 here because we want the coordation
            # to be atom-dependent (best we can do is oxidation-state dependent here)
            sitefingerprint_method = CrystalNNFingerprint.from_preset(
                "ops", distance_cutoffs=None, x_diff_weight=3
            )

        # now that we made the sitefingerprint_method, we can input it into the
        # structurefingerprint_method which finishes up the featurizer
        self.featurizer = SiteStatsFingerprint(
            sitefingerprint_method, stats=stat_options
        )

        # for this specific featurizer, you're supposed to use the .fit() function
        # with an example structure but really, all that does is get a list of
        # unique elements - so we can make that with a composition object
        self.featurizer.elements_ = numpy.array(
            [element.symbol for element in composition.elements]
        )

        # check if we were given a list of pymatgen structures.
        if isinstance(structure_pool, list):
            # set this variable as none to help with some warnings methods such
            # as the update_fingerprint_database
            self.structure_pool_queryset = "local_only"
            # If so, we generate the fingerprint for each of the initial input structures
            # We convert this to a numpy array for speed improvement at later stages
            # OPTIMIZE: this can be slow and I should support parallel featurization
            self.fingerprint_database = numpy.array(
                [
                    self.featurizer.featurize(structure)
                    for structure in tqdm(structure_pool)
                ]
            )

        # otherwise we have a queryset that should be used to populate the
        # fingerprint database
        else:
            # we store the queryset as an attribute because we may want to
            # update the structure pool later on.
            self.structure_pool_queryset = structure_pool
            self.fingerprint_database = numpy.array([])
            # we also keep a log of the last update so we only grab new structures
            # each time we update the database. To start, we set this as the
            # eariest possible date, which tells our update_fingerprint_database
            # method to include ALL structures
            self.last_update = timezone.make_aware(
                timezone.datetime.min, timezone.get_default_timezone()
            )
            self.update_fingerprint_database()

    def check_structure(self, structure, tolerance=0.001):
        # OPTIMIZE: my choice of tolerance is based on my evolutionary search.
        # However, "clustering" after evolutionary search suggests that something
        # like 0.5 might be a better choice for finding a truly unique structure

        # make the fingerprint for this structure and make into a numpy array for speed
        fingerprint1 = numpy.array(self.featurizer.featurize(structure))

        # We now want to get the distance of this fingerprint relative to all others.
        # If the distance is within the specified tolerance, then the structures
        # are too similar - and we return  for a failure.
        for fingerprint2 in self.fingerprint_database:
            distance = numpy.linalg.norm(fingerprint1 - fingerprint2)
            if distance < tolerance:
                # we can end the whole function as soon as one structure is deemed too similar
                return False
            else:
                continue
        # If we make it through all structures and no distance is below the
        # tolerance, we have a new and unique structure!

        # add this new structure to the database if it was requested.
        if self.add_unique_to_pool:
            self._add_fingerprint_to_database(fingerprint1)

        # Return that we were successful
        return True

    def update_fingerprint_database(self):

        if self.structure_pool_queryset == "local_only":
            raise Exception(
                "This method should only be used when your structure pool"
                " is based on a Simmate database table!"
            )

        # BUG: what if a structure was added to the table WHILE I was running the
        # last query? This sets up a race condition with new structures and my
        # last_update timestamp. To avoid this bug, I immediately update our
        # timestamp, rather than doing it after the query
        last_update_safe = self.last_update
        self.last_update = timezone.now()

        # now grab all the new structures!
        new_structures = (
            self.structure_pool_queryset.filter(created_at__gte=last_update_safe)
            .only("structure_string")
            .to_toolkit()
        )

        # If there aren't any new structures, just exit without printing the
        # message and tqdm progress bar below
        if not new_structures:
            return

        print(
            f"Found {len(new_structures)} new structures for the fingerprint database."
        )

        # calculate each fingerprint and add it to the database
        for structure in tqdm(new_structures):
            fingerprint = numpy.array(self.featurizer.featurize(structure))
            self._add_fingerprint_to_database(fingerprint)

    def _add_fingerprint_to_database(self, fingerprint):
        if self.fingerprint_database.size == 0:
            self.fingerprint_database = numpy.array([fingerprint])
        else:
            self.fingerprint_database = numpy.append(
                self.fingerprint_database, [fingerprint], axis=0
            )



# -*- coding: utf-8 -*-

from simmate.workflow_engine import ErrorHandler

from simmate.calculators.vasp.error_handlers import TetrahedronMesh, IncorrectShift


class Tetirr(ErrorHandler):
    """
    This handler addresses an issue that is a combination of the TetrahedronMesh
    and IncorrectShift errors.
    """

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = ["Routine TETIRR needs special values"]

    def correct(self, directory: str) -> str:

        # apply the fixes uses these other error handlers
        fixes_1 = TetrahedronMesh().correct(directory)
        fixes_2 = IncorrectShift().correct(directory)

        # combine the output messages and return
        return f"{fixes_1} and {fixes_2}"



# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod

from dask.distributed import get_client


class StructureCreator(ABC):
    def __init__(self):
        """
        This is where you store all settings for the creation and do the setup.
        For example, you would store the target composition here and/or
        specify the fixed volume to use. Many times, there are 'setup' steps
        that require a lot of overhead which you don't want to redo for each
        new creation. For example, you don't want to find all combinations
        of allowed wyckoff sites for a specific composition each time you make
        a new structure, so you would find all valid combinations here in the
        __init__ so you don't have to repeatedly find combos below in create()
        """
        pass

    @abstractmethod
    def create_structure(self, spacegroup=None):
        """
        For all different structure creators, there should be a method of
        .new_structure() that makes the new object. The main rule for this
        function is that it must return a single pymatgen structure object.
        How that structure object is created is up to each method!

        Below is some example code for how Structures are typically made. How
        the spacegroup/lattice/species/coords is determined is up to each
        methodand not shown.

        # import module (do this outside the class functions)
        from simmate.toolkit import Structure

        # run your method that makes the lattice/species/coords or uses the
        # inputs from above.
        ...

        # for new_structure method that doesnt use symmetry or wyckoff sites
        sg = 1 # this is never used but just for bookkeeping
        structure = Structure(lattice = lattice,
                              species = species,
                              coords = coords)

        # for new_structure methods that depend on symmetry and wyckoff sites
        # If you give the sites in the asym unit, they will be replicated
        # elsewhere in the unitcell using spacegroup symmetry operations.
        structure = Structure.from_spacegroup(sg = spacegroup,
                                              lattice = lattice,
                                              species = species,
                                              coords = coords)
        """
        raise NotImplementedError

    def create_many_structures(self, n, spacegroup=None):
        """
        This is a convience function to create many structures in parallel.
        In most cases, you don't want to redefine this function.
        It is assumed that calling create_structure() repeatedly is the
        fastest method to make new structures, so that is the function
        parallelized. In some cases (such as the USPEX creator), there is a
        lot of overhead with a single call to create_structure(), so it may
        make more sense to have the main structure creation code in this
        function. See the USPEXStructure Creator class for an example of this.

        Make sure you have a Dask cluster setup as a global variable!
        Here's how you should do that...
            from dask.distributed import Client
            client = Client(processes=False)
        """
        # USING DASK FUTURES TO PARALLELIZE

        # grab the dask cluster
        #!!! Maybe give the client as an optional input? And if none is
        #!!! given or detected, start up one automatically...?
        try:
            client = get_client()
        except ValueError:
            print("Set up a Dask cluster before trying to run this!")
            return False
        # launch create_structure(spacegroup) iteratively and in parallel
        # The pure=False indicates that we will have different answers, even for the same input
        futures = client.map(self.create_structure, [spacegroup] * n, pure=False)
        # wait for all of the calls to finish and grab the results
        results = client.gather(futures)  #!!! faster for dask
        # we now have a list of structures of size n and can return them
        return results

    def update_data(self):
        """
        TO-DO
        This is a function to update settings. For example, I could use this
        to update the self.lattice_creator method to a more accurate volume.
        I can run machine learning code here before updating too.
        """
        pass



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.base import VaspTask
from simmate.calculators.vasp.inputs.potcar_mappings import (
    PBE_ELEMENT_MAPPINGS_LOW_QUALITY,
)
from simmate.calculators.vasp.error_handlers import (
    TetrahedronMesh,
    RotationMatrix,
    Brmix,
    Zpotrf,
    SubspaceMatrix,
    IncorrectShift,
    RealOptlay,
    Tetirr,
    RotationNonIntMatrix,
    LongVector,
    TripleProduct,
    Pricel,
    Brions,
    Zbrent,
    InsufficientBands,
    Pssyevx,
    Eddrmm,
    Edddav,
    Edwav,
    Zheev,
    ElfKpar,
    Rhosyg,
    Posmap,
    PointGroup,
    SymprecNoise,
    IncorrectSmearing,
    MeshSymmetry,
    Unconverged,
    NonConverging,
    Potim,
    PositiveEnergy,
    Frozen,
    LargeSigma,
    Walltime,
)


class MITRelaxation(VaspTask):
    """
    This task is a reimplementation of pymatgen's
    [MITRelaxSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MITRelaxSet).

    Runs a VASP geometry optimization using MIT Project settings, where the
    MIT Project is the precursor to the Materials Project.

    From their documention in pymatgen:
    ```
    Default VASP settings for calculations in the pre-cursor MIT project to
    the Materials Project. Reasonably robust, but selected PSP are generally the
    ones with fewer electrons and convergence criteria is less tight. This was
    used in an era whether computational power is more limited.
    Nevertheless, this is still a good starting point for extremely expensive
    methods.
    ```
    """

    functional = "PBE"
    potcar_mappings = PBE_ELEMENT_MAPPINGS_LOW_QUALITY

    incar = dict(
        ALGO="Fast",
        EDIFF=1.0e-05,
        ENCUT=520,
        IBRION=2,
        ICHARG=1,
        ISIF=3,
        ISPIN=2,
        ISYM=0,
        LORBIT=11,
        LREAL="Auto",
        LWAVE=False,
        NELM=200,
        NELMIN=6,
        NSW=99,
        PREC="Accurate",
        ISMEAR=-5,
        SIGMA=0.05,
        KSPACING=0.5,  # !!! This is where we are different from pymatgen right now
        MAGMOM__smart_magmom={
            "default": 0.6,  # note this is different from the VASP default
            "Ce": 5,
            "Ce3+": 1,
            "Co": 0.6,
            "Co3+": 0.6,
            "Co4+": 1,
            "Cr": 5,
            "Dy3+": 5,
            "Er3+": 3,
            "Eu": 10,
            "Eu2+": 7,
            "Eu3+": 6,
            "Fe": 5,
            "Gd3+": 7,
            "Ho3+": 4,
            "La3+": 0.6,
            "Lu3+": 0.6,
            "Mn": 5,
            "Mn3+": 4,
            "Mn4+": 3,
            "Mo": 5,
            "Nd3+": 3,
            "Ni": 5,
            "Pm3+": 4,
            "Pr3+": 2,
            "Sm3+": 5,
            "Tb3+": 6,
            "Tm3+": 2,
            "V": 5,
            "W": 5,
            "Yb3+": 1,
        },
        LMAXMIX__smart_lmaxmix=True,
        multiple_keywords__smart_ldau=dict(
            LDAU__auto=True,
            LDAUTYPE=2,
            LDAUPRINT=1,
            LDAUJ={},  # pymatgen sets these, but they're all default values anyways
            LDAUL={
                "F": {
                    "Ag": 2,
                    "Co": 2,
                    "Cr": 2,
                    "Cu": 2,
                    "Fe": 2,
                    "Mn": 2,
                    "Mo": 2,
                    "Nb": 2,
                    "Ni": 2,
                    "Re": 2,
                    "Ta": 2,
                    "V": 2,
                    "W": 2,
                },
                "O": {
                    "Ag": 2,
                    "Co": 2,
                    "Cr": 2,
                    "Cu": 2,
                    "Fe": 2,
                    "Mn": 2,
                    "Mo": 2,
                    "Nb": 2,
                    "Ni": 2,
                    "Re": 2,
                    "Ta": 2,
                    "V": 2,
                    "W": 2,
                },
                "S": {
                    "Fe": 2,
                    "Mn": 2.5,
                },
            },
            LDAUU={
                "F": {
                    "Ag": 1.5,
                    "Co": 3.4,
                    "Cr": 3.5,
                    "Cu": 4,
                    "Fe": 4.0,
                    "Mn": 3.9,
                    "Mo": 4.38,
                    "Nb": 1.5,
                    "Ni": 6,
                    "Re": 2,
                    "Ta": 2,
                    "V": 3.1,
                    "W": 4.0,
                },
                "O": {
                    "Ag": 1.5,
                    "Co": 3.4,
                    "Cr": 3.5,
                    "Cu": 4,
                    "Fe": 4.0,
                    "Mn": 3.9,
                    "Mo": 4.38,
                    "Nb": 1.5,
                    "Ni": 6,
                    "Re": 2,
                    "Ta": 2,
                    "V": 3.1,
                    "W": 4.0,
                },
                "S": {
                    "Fe": 1.9,
                    "Mn": 2.5,
                },
            },
        ),
    )

    error_handlers = [
        TetrahedronMesh(),
        RotationMatrix(),
        Brmix(),
        Zpotrf(),
        SubspaceMatrix(),
        IncorrectShift(),
        RealOptlay(),
        Tetirr(),
        RotationNonIntMatrix(),
        LongVector(),
        TripleProduct(),
        Pricel(),
        Brions(),
        Zbrent(),
        InsufficientBands(),
        Pssyevx(),
        Eddrmm(),
        Edddav(),
        Edwav(),
        Zheev(),
        ElfKpar(),
        Rhosyg(),
        Posmap(),
        PointGroup(),
        SymprecNoise(),
        IncorrectSmearing(),
        MeshSymmetry(),
        Unconverged(),
        NonConverging(),
        Potim(),
        PositiveEnergy(),
        Frozen(),
        LargeSigma(),
        Walltime(),
    ]



# -*- coding: utf-8 -*-

from textwrap import dedent

from django import template
from django.utils.safestring import mark_safe

from pdoc.render_helpers import to_html


# We need a registration instance in order to configure everything with Django
register = template.Library()


@register.filter(name="markdown")
def markdown_to_html(markdown_str):
    """
    Converts markdown to rendered HTML using utilities from pdoc. This is especially
    useful for rendering python docstrings.
    """

    # Docstrings often have the entire string indented a number of times.
    # We need to strip those idents away to render properly
    markdown_cleaned = dedent(markdown_str)

    final_html = to_html(markdown_cleaned)

    # Because we added new html to our script, we need to have Django check it
    # ensure it safe before returning. Read more about this here:
    # https://docs.djangoproject.com/en/3.2/howto/custom-template-tags/#filters-and-auto-escaping
    return mark_safe(final_html)



# -*- coding: utf-8 -*-

import os

import django
from django.conf import settings


def setup_django():
    """
    Configures django settings. This method does not need to be called directly
    because importing this module does so automatically.

    The first time this function is used can take >1 seconds because many things
    are imported from the start. Thereafter, testing shows that only <200 ns
    are used when settings are already configured.
    """

    # OPTIMIZE: I should have setup_select() function that only configures
    # "select" apps to help with setup speed.

    # see if django has already been configured. If so, just exit this function.
    if settings.configured:
        return
    # BUG: originally I used the code below, but it didn't work with Prefect+Dask:
    #   if "DJANGO_SETTINGS_MODULE" in os.environ:

    # The code below is the equiv of running 'python manage.py shell'
    # This sets up all of django for its full use. You can being importing
    # models after this is setup.
    os.environ.setdefault(
        "DJANGO_SETTINGS_MODULE", "simmate.configuration.django.settings"
    )
    django.setup()


# When I import this module, it automatically configures django for us, including
# connecting to the database(s). Without this file, I would instead need these two
#  lines in every single file before I import models:
#   from simmate.configuration.django import setup
#   setup_django() # ensures setup
# With this init, I instead only need to have one line:
#   from simmate.configuration.django import setup
# Thus this "module" is more of a script.
setup_django()



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.relaxation import (
    MatVirtualLabGrainBoundaryRelaxation as MVLGBRelax,
)


class MatVirtualLabSlabRelaxation(MVLGBRelax):
    """
    This task is a reimplementation of pymatgen's
    [MVLGBSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MVLGBSet)
    with slab_mode=True.
    """

    # The settings used for this calculation are based on the MITRelaxation, but
    # we are updating/adding new settings here.
    incar = MVLGBRelax.incar.copy()
    incar.update(
        dict(
            ISIF=2,
            NELMIN=8,
        )
    )



# -*- coding: utf-8 -*-

# OPTIMIZE: using caching improves speed significantly at the cost of memory. I
# need to see which is preferable in higher-level tests.
from functools import cached_property

import numpy
from numba import njit as numba_speedup


class Lattice:
    """
    This class is for 3D crystal lattices, which are represented by a 3x3 matrix.
    This class provides a bunch of tools to pull out common values like vector
    lengths and angles, and also allows for lattice conversions.

    For some basic formulas and introduction, see here:
        http://gisaxs.com/index.php/Unit_cell
        http://gisaxs.com/index.php/Lattices
    """

    def __init__(self, matrix):
        """
        Create a lattice from a 3x3 matrix. Each vector is defined via xyz on
        one row of the matrix. For example, a cubic lattice where all vectors
        are 1 Angstrom and 90deg from eachother would be defined as
            matrix = [[1, 0, 0],
                      [0, 1, 0],
                      [0, 0, 1]]
        """

        # Convert the matrix to a numpy array and store it
        self.matrix = numpy.array(matrix)

    @cached_property
    def lengths(self):
        """
        Gives the lengths of the lattice as (a, b, c).
        """
        # The easiest way to do this is with numpy via...
        #   lengths = numpy.linalg.norm(matrix, axis=1)
        # However, we instead write a super-optimized numba function to make this
        # method really fast. This is defined below as a static method.
        # The length of a 3D vector (xyz) is defined by...
        #   length = sqrt(a**2 + b**2 + c**2)
        # and we are doing this for all three vectors at once.
        return self._lengths_fast(self.matrix)

    @staticmethod
    @numba_speedup(cache=True)
    def _lengths_fast(matrix):

        # NOTE: users should not call this method! Use lattice.lengths instead

        # OPTIMIZE: is there an alternative way to write this that numba will
        # run faster?

        #   numpy.linalg.norm(matrix, axis=1) --> doesn't work with numba
        return numpy.sqrt(numpy.sum(matrix**2, axis=1))

    @property
    def a(self):
        """
        The length of the lattice vector "a" (in Angstroms)
        """
        return self.lengths[0]

    @property
    def b(self):
        """
        The length of the lattice vector "b" (in Angstroms)
        """
        return self.lengths[1]

    @property
    def c(self):
        """
        The length of the lattice vector "c" (in Angstroms)
        """
        return self.lengths[2]

    @cached_property
    def angles(self):
        """
        The angles of the lattice as (alpha, beta, gamma).
        """
        # The angle between two 3D vectors (xyz) is defined by...
        #   angle_radians = arccos(dot(vector1, vector2)/(vector1.length*vector2.length))
        # And alpha, beta, gamma are just the angles between the b&c, a&c, and a&b
        # vectors, respectively. We cacluate all of these at once here. We also
        # write a super-optimized numba function to make this method really fast.
        # This is defined below as a static method.
        return self._angles_fast(self.matrix, self.lengths)

    @staticmethod
    @numba_speedup(cache=True)
    def _angles_fast(matrix, lengths):

        # NOTE: users should not call this method! Use lattice.angles instead

        # OPTIMIZE: is there an alternative way to write this that numba will
        # run faster?

        # keep a list of the angles
        angles = []

        # this establishes the three angles we want to calculate
        #   alpha --> angle between b and c
        #   beta --> angle between a and c
        #   gamma --> angle between a and b
        # So the numbers below are indexes of the matrix!
        # (for example, vector b has index 1)
        vector_pairs = [(1, 2), (0, 2), (0, 1)]

        # The angle between two 3D vectors (xyz) is defined by...
        #   angle_radians = arccos(dot(vector1, vector2)/(vector1.length*vector2.length))
        for vector_index_1, vector_index_2 in vector_pairs:
            # calculate single angle using formula
            unit_vector_1 = matrix[vector_index_1] / lengths[vector_index_1]
            unit_vector_2 = matrix[vector_index_2] / lengths[vector_index_2]
            dot_product = numpy.dot(unit_vector_1, unit_vector_2)
            angle = numpy.arccos(dot_product)
            # convert to degrees before saving
            angles.append(numpy.degrees(angle))

        # return the list as a numpy array
        return numpy.array(angles)

    @property
    def alpha(self):
        """
        The angle between the lattice vectors b and c (in degrees).
        """
        return self.angles[0]

    @property
    def beta(self):
        """
        The angle between the lattice vectors a and c (in degrees).
        """
        return self.angles[1]

    @property
    def gamma(self):
        """
        The angle between the lattice vectors a and b (in degrees).
        """
        return self.angles[2]



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.base import VaspTask
from simmate.calculators.vasp.inputs.potcar_mappings import (
    PBE_ELEMENT_MAPPINGS,
)
from simmate.calculators.vasp.error_handlers import (
    TetrahedronMesh,
    RotationMatrix,
    Brmix,
    Zpotrf,
    SubspaceMatrix,
    IncorrectShift,
    RealOptlay,
    Tetirr,
    RotationNonIntMatrix,
    LongVector,
    TripleProduct,
    Pricel,
    Brions,
    Zbrent,
    InsufficientBands,
    Pssyevx,
    Eddrmm,
    Edddav,
    Edwav,
    Zheev,
    ElfKpar,
    Rhosyg,
    Posmap,
    PointGroup,
    SymprecNoise,
    IncorrectSmearing,
    MeshSymmetry,
    Unconverged,
    NonConverging,
    Potim,
    PositiveEnergy,
    Frozen,
    LargeSigma,
    Walltime,
)


class MatprojHSERelaxation(VaspTask):
    """
    This task is a reimplementation of pymatgen's
    [MPHSERelaxSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPHSERelaxSet).

    Runs a VASP relaxation calculation using Materials Project HSE settings.
    """

    functional = "PBE"
    potcar_mappings = PBE_ELEMENT_MAPPINGS

    incar = dict(
        ALGO="All",
        EDIFF__per_atom=5.0e-05,
        ENCUT=520,
        HFSCREEN=0.2,
        IBRION=2,
        ICHARG=1,
        ISIF=3,
        ISMEAR=0,
        ISPIN=2,
        LHFCALC=True,
        LORBIT=11,
        LREAL="Auto",
        LWAVE=False,
        NELM=100,
        NSW=99,
        PREC="Accurate",
        PRECFOCK="Fast",
        SIGMA=0.05,
        KSPACING=0.4,  # !!! This is where we are different from pymatgen right now
        MAGMOM__smart_magmom={
            "default": 0.6,  # note this is different from the VASP default
            "Ce": 5,
            "Ce3+": 1,
            "Co": 0.6,
            "Co3+": 0.6,
            "Co4+": 1,
            "Cr": 5,
            "Dy3+": 5,
            "Er3+": 3,
            "Eu": 10,
            "Eu2+": 7,
            "Eu3+": 6,
            "Fe": 5,
            "Gd3+": 7,
            "Ho3+": 4,
            "La3+": 0.6,
            "Lu3+": 0.6,
            "Mn": 5,
            "Mn3+": 4,
            "Mn4+": 3,
            "Mo": 5,
            "Nd3+": 3,
            "Ni": 5,
            "Pm3+": 4,
            "Pr3+": 2,
            "Sm3+": 5,
            "Tb3+": 6,
            "Tm3+": 2,
            "V": 5,
            "W": 5,
            "Yb3+": 1,
        },
        LMAXMIX__smart_lmaxmix=True,
    )

    error_handlers = [
        TetrahedronMesh(),
        RotationMatrix(),
        Brmix(),
        Zpotrf(),
        SubspaceMatrix(),
        IncorrectShift(),
        RealOptlay(),
        Tetirr(),
        RotationNonIntMatrix(),
        LongVector(),
        TripleProduct(),
        Pricel(),
        Brions(),
        Zbrent(),
        InsufficientBands(),
        Pssyevx(),
        Eddrmm(),
        Edddav(),
        Edwav(),
        Zheev(),
        ElfKpar(),
        Rhosyg(),
        Posmap(),
        PointGroup(),
        SymprecNoise(),
        IncorrectSmearing(),
        MeshSymmetry(),
        Unconverged(),
        NonConverging(),
        Potim(),
        PositiveEnergy(),
        Frozen(),
        LargeSigma(),
        Walltime(),
    ]
